from moai.utils.arguments import assert_path
from moai.data.datasets.common.image2d import (
    load_color_image,
    load_binary_image,
)

import numpy as np
import torch
import os
import typing
import logging
import json

log = logging.getLogger(__name__)

__all__ = ['SSP3D']

#NOTE: Human3.6M body structure: {
    # 0: 'Pelvis',
    # 1: 'RHip',
    # 2: 'RKnee',
    # 3: 'RAnkle',
    # 4: 'LHip',
    # 5: 'LKnee',
    # 6: 'LAnkle',
    # 7: 'Spine1',
    # 8: 'Neck',
    # 9: 'Head',
    # 10: 'Site',
    # 11: 'LShoulder',
    # 12: 'LElbow',
    # 13: 'LWrist',
    # 14: 'RShoulder',
    # 15: 'RElbow',
    # 16: 'RWrist,
# };

#NOTE: COCO17 body structure: {
    # 0: 'Nose',
    # 1: 'REye',
    # 2: 'LEye',
    # 3: 'REar',
    # 4: 'LEar',
    # 5: 'RShoulder',
    # 6: 'LShoulder',
    # 7: 'RElbow',
    # 8: 'LElbow',
    # 9: 'RWrist',
    # 10: 'LWrist',
    # 11: 'RHip',
    # 12: 'LHip',
    # 13: 'RKnee',
    # 14: 'LKnee',
    # 15: 'RAnkle',
    # 16: 'LAnkle,
# };

class SSP3D(torch.utils.data.Dataset):
    def __init__(self,
        root:           str,
        betas:          int=10,
        with_openpose:  bool=False,
    ) -> None:
        super().__init__()
        assert_path(log, 'EHF root path', root)
        labels = np.load(os.path.join(root, 'labels.npz'))
        self.root = root
        self.betas = betas
        self.with_openpose = with_openpose
        for k, v in labels.items():
            setattr(self, k, torch.from_numpy(v) if np.issubdtype(v.dtype, np.number) else v)

    def __len__(self) -> int:
        return len(self.fnames)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        data = {
            'color': load_color_image(os.path.join(self.root, 'images', self.fnames[index])),
            'silhouette': load_binary_image(os.path.join(self.root, 'silhouettes', self.fnames[index])),
            'shape': torch.cat([self.shapes[index], torch.zeros(self.betas-10)], dim=-1) if self.betas > 10 else self.shapes[index],
            'global_orientation': self.poses[index][:3],
            'pose': self.poses[index][3:],
            'keypoints': self.joints2D[index][:, :2],
            'confidence': self.joints2D[index][:, 2:3],
            'camera_translation': self.cam_trans[index],
        }
        if self.with_openpose:
            filename = os.path.basename(self.fnames[index])
            kpts_all = self._read_keypoints(os.path.join(
                self.root, 'keypoints', filename.replace(".png", "_keypoints.json")
            ))['keypoints']
            coco17 = np.array([0, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16])
            coco25 = np.array([0, 2, 5, 3, 6, 4, 7, 9, 12, 10, 13, 11, 14])
            selected = min(kpts_all, 
                key=lambda kpts: np.linalg.norm(kpts[coco25][:, :2]-data['keypoints'][coco17]).mean()
            )
            data['openpose'] = {
                'keypoints': selected[:, :2],
                'confidence': selected[:, 2:3],
            }        
        return data

    #TODO: refactor to use the same func (i.e. from openpose.py)
    def _read_keypoints(self,
        filename: str,
    ) -> typing.Dict[str, torch.Tensor]:
        with open(filename) as keypoint_file:
            data = json.load(keypoint_file)
        keypoints, gender_pd, gender_gt = [], [], []
        for person in data['people']:
            body = np.array(person['pose_keypoints_2d'], dtype=np.float32)
            body = body.reshape([-1, 3])            
            left_hand = np.array(person['hand_left_keypoints_2d'], dtype=np.float32).reshape([-1, 3])
            right_hand = np.array(person['hand_right_keypoints_2d'], dtype=np.float32).reshape([-1, 3])
            body = np.concatenate([body, left_hand, right_hand], axis=0)            
            face = np.array(person['face_keypoints_2d'], dtype=np.float32).reshape([-1, 3])[17: 17 + 51, :]
            contour_keyps = np.array([], dtype=body.dtype).reshape(0, 3)
            body = np.concatenate([body, face, contour_keyps], axis=0)
            gender_pd.append(person.get('gender_pd', None))
            gender_gt.append(person.get('gender_gt', None))
            keypoints.append(torch.from_numpy(body))
        return {
            'keypoints': keypoints,
            'gender_pd': gender_pd,
            'gender_gt': gender_gt, 
        }


from moai.engine.callbacks import EngineCallbacks
import moai.engine.lightning.play as miplay

import sys
import hydra
import omegaconf.omegaconf
import logging

log = logging.getLogger(__name__)

def play(cfg):
    hydra.utils.log.debug(f"Configuration:\n{omegaconf.OmegaConf.to_yaml(cfg, resolve=True)}")
    engine = hydra.utils.instantiate(cfg.engine)
    model = miplay.Presenter(
        **(cfg.model if hasattr(cfg, "model") else {"monads": None}) ,
        data=cfg.data,
        visualization=cfg.visualization,
        statistics=cfg.statistics if hasattr(cfg, "statistics") else None,
    )
    player = hydra.utils.instantiate(cfg.player, 
        engine_callbacks=EngineCallbacks(
            visualization=model.visualization,
            statistics=model.statistics
        )
    )
    log.info("Dataset showcase started.")
    player.play(model)
    log.info("Dataset showcase completed.")

if __name__ == "__main__":  
    # os.environ['HYDRA_FULL_ERROR'] = '1'
    config_filename = sys.argv.pop(1) #TODO: argparser integration?   
    sys.argv.append("hydra.run.dir=actions/${hydra.job.name}/${now:%Y-%m-%d}/${now:%H-%M-%S}-${experiment.name}")
    play = hydra.main(config_path="conf", config_name=config_filename)(play)
    play()


import pytorch_lightning
import typing
import omegaconf.omegaconf

__all__ = ["NoOp"]

class NoOp(pytorch_lightning.loggers.base.LightningLoggerBase):
    def __init__(self,
        loggers: omegaconf.DictConfig=None,
        **kwargs
    ):
        super(NoOp, self).__init__()

    @property
    def name(self):
        return "NoLogging"
        
    @pytorch_lightning.loggers.base.rank_zero_only
    def log_metrics(self, 
        metrics: typing.Dict[str, typing.Any],
        step: int
    ):
       pass

    @pytorch_lightning.loggers.base.rank_zero_only
    def log_hyperparams(self,
        params: typing.Dict[str, typing.Any], #TODO: or namespace object ?
    ):
        pass

    @pytorch_lightning.loggers.base.rank_zero_only
    def save(self):
        pass

    @pytorch_lightning.loggers.base.rank_zero_only
    def finalize(self, 
        status: str
    ):
        pass

    @property
    def rank(self):
        return 0

    @rank.setter
    def rank(self, value: int):
        pass

    @property
    def version(self):
        return 0

    @property
    def experiment(self) -> typing.Any:
        return self.name


import torch
import functools
import typing

__all__ = ['HypergradSGD']

#NOTE: modified from https://github.com/wbaek/torchskeleton/tree/master/skeleton/optim
#NOTE: code from https://github.com/gbaydin/hypergradient-descent/blob/master/hypergrad/sgd_hd.py

class HypergradSGD(torch.optim.Optimizer):
    r""" Implements the Hypergradient Descent Algorithm.

    - **Paper**: [ONLINE LEARNING RATE ADAPTATION WITH HYPERGRADIENT DESCENT](https://openreview.net/pdf?id=BkrsAzWAb)
    - **Implementation**: [GitHub @ gbaydin](https://github.com/gbaydin/hypergradient-descent/)

    """
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        #lr: torch.optim.optimizer.required,
        lr=None,
        momentum: float=0,
        dampening: float=0,
        weight_decay: float=0,
        nesterov: bool=False,
        hypergrad_lr: float=0,
        hypergrad_momentum: float=0,
        min_lr: float=1e-6
    ):
        #if lr is not torch.optim.optimizer.required and lr < 0.0:
        if lr is not None and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if momentum < 0.0:
            raise ValueError("Invalid momentum value: {}".format(momentum))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))

        defaults = dict(lr=lr, momentum=momentum, dampening=dampening,
                        weight_decay=weight_decay, nesterov=nesterov,
                        hypergrad_lr=hypergrad_lr, hypergrad_momentum=hypergrad_momentum, min_lr=min_lr)
        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError("Nesterov momentum requires a momentum and zero dampening")
        super(HypergradSGD, self).__init__(params, defaults)

        for group in self.param_groups:
            group['prev_grads'] = [torch.zeros_like(p, requires_grad=False) for p in group['params']]
            group['lr_per_layers'] = [group['lr'] for p in group['params']]

    def __setstate__(self, state):
        super(HypergradSGD, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('nesterov', False)

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            dampening = group['dampening']
            nesterov = group['nesterov']

            hypergrad_lr = group['hypergrad_lr']
            hypergrad_momentum = group['hypergrad_momentum']
            min_lr = group['min_lr']

            lr_per_layers = group['lr_per_layers']

            for index, (p, d_p_prev) in enumerate(zip(group['params'], group['prev_grads'])):
                if p.grad is None:
                    continue

                d_p = p.grad.data
                h = torch.dot(d_p.view(-1), d_p_prev.view(-1))
                # group['prev_grads'][index] = d_p.clone().detach().requires_grad_(False)

                if hypergrad_momentum != 0:
                    param_state = self.state[p]
                    if 'momentum_buffer_h' not in param_state:
                        buf_h = param_state['momentum_buffer_h'] = torch.clone(h).detach()
                    else:
                        buf_h = param_state['momentum_buffer_h']
                        buf_h.mul_(hypergrad_momentum).add_(1, h)
                    h = buf_h

                lr = lr_per_layers[index]
                lr = max(min_lr, lr + (hypergrad_lr * h))
                lr_per_layers[index] = lr

                if momentum != 0:
                    param_state = self.state[p]
                    if 'momentum_buffer' not in param_state:
                        buf = param_state['momentum_buffer'] = torch.clone(d_p).detach()
                    else:
                        buf = param_state['momentum_buffer']
                        buf.mul_(momentum).add_(1 - dampening, d_p)
                    if nesterov:
                        d_p = d_p.add(momentum, buf)
                    else:
                        d_p = buf

                group['prev_grads'][index] = d_p.clone().detach().requires_grad_(False)
                p.data.add_(-lr, d_p)

                # Apply decoupled weight decay
                if weight_decay != 0:
                    p.data.add_(-lr, weight_decay)

            group['lr'] = sum(group['lr_per_layers']) / len(group['lr_per_layers'])

        return loss



import torch

__all__ = ['SGDW']

#NOTE: modified from https://github.com/wbaek/torchskeleton/tree/master/skeleton/optim

class SGDW(torch.optim.SGD):
    """Implements SGD with weight decay

    - **Paper**: [Decoupled Weight Decay Regularization](https://arxiv.org/abs/1711.05101)
    - **Implementation**: [GitHub @ wbaek](https://github.com/wbaek/torchskeleton/tree/master/skeleton/optim)
    
    """
    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            dampening = group['dampening']
            nesterov = group['nesterov']

            for p in group['params']:
                if p.grad is None:
                    continue
                d_p = p.grad.data
                if momentum != 0:
                    param_state = self.state[p]
                    if 'momentum_buffer' not in param_state:
                        buf = param_state['momentum_buffer'] = torch.zeros_like(p.data)
                        buf.mul_(momentum).add_(d_p)
                    else:
                        buf = param_state['momentum_buffer']
                        buf.mul_(momentum).add_(1 - dampening, d_p)
                    if nesterov:
                        d_p = d_p.add(momentum, buf)
                    else:
                        d_p = buf

                # decoupled weight decay
                if weight_decay != 0:
                    d_p.add_(weight_decay, p.data)

                p.data.add_(-group['lr'], d_p)

        return loss



import moai.nn.utils as miu

import torch
import logging
import typing

log = logging.getLogger(__name__)

__all__ = [
    "make_downsample"
]

__DOWNSAMPLE_FACTORY__ = {
    "none":                 torch.nn.Identity,
    "identity":             torch.nn.Identity,
    "in_block":             torch.nn.Identity, #NOTE: needed for resnet
    "maxpool2d":            torch.nn.MaxPool2d,
    "avgpool2d":            torch.nn.AvgPool2d,
}
        
def _update_downsample_op(name: str, type: typing.Type):
    if name not in __DOWNSAMPLE_FACTORY__.keys():
        __DOWNSAMPLE_FACTORY__.update({name: type})
    else:
        log.error(f"Trying to add an already existing key {name} in the convolution operation factory.")

def make_downsample(
    downscale_type: str,
    features: int,
    kernel_size: int=3,
    stride: int=2,
    **kwargs
) -> torch.nn.Module:
    if downscale_type in __DOWNSAMPLE_FACTORY__.keys():
        return miu.instantiate(__DOWNSAMPLE_FACTORY__[downscale_type], 
        **{
            **locals(),
            **kwargs 
        })
    else:
        log.error(f"Downscale type ({downscale_type}) not found.")

import moai.nn.sampling.spatial.downsample.conv as midsc

if "conv2d" not in __DOWNSAMPLE_FACTORY__.keys():
    _update_downsample_op("conv2d", midsc.StridedConv2d)

del midsc

import moai.nn.sampling.spatial.downsample.interpolate as midsi

if "downsample2d" not in __DOWNSAMPLE_FACTORY__.keys():
    _update_downsample_op("downsample2d", midsi.Downsample2d)

del midsi

import moai.nn.sampling.spatial.downsample.antialiased as midsaa

if "maxpool2d_aa" not in __DOWNSAMPLE_FACTORY__.keys():
    _update_downsample_op("maxpool2d_aa", midsaa.AntialiasedMaxPool2d)
    _update_downsample_op("conv2d_aa", midsaa.AntialiasedStridedConv2d)

del midsaa

import moai.nn.sampling.spatial.downsample.adaptive as midsad

if "adaptive" not in __DOWNSAMPLE_FACTORY__.keys():
    _update_downsample_op("adaptive", midsad.Adaptive)
    _update_downsample_op("adaptive1d", midsad.Adaptive1d)
    _update_downsample_op("adaptive2d", midsad.Adaptive2d)
    _update_downsample_op("adaptive3d", midsad.Adaptive3d)
    _update_downsample_op("adaptive1d_max", midsad.AdaptiveMaxPool1d)
    _update_downsample_op("adaptive2d_max", midsad.AdaptiveMaxPool2d)
    _update_downsample_op("adaptive3d_max", midsad.AdaptiveMaxPool3d)
    _update_downsample_op("adaptive1d_avg", midsad.AdaptiveAvgPool1d)
    _update_downsample_op("adaptive2d_avg", midsad.AdaptiveAvgPool2d)
    _update_downsample_op("adaptive3d_avg", midsad.AdaptiveAvgPool3d)

del midsad


import moai.networks.lightning as minet
import moai.utils.parsing.rtp as mirtp

import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging
import inspect

log = logging.getLogger(__name__)

__all__ = ['Null']

from moai.monads.execution.cascade import _create_accessor

class Null(minet.FeedForward):
    def __init__(self,
        data:           omegaconf.DictConfig=None,
        parameters:     omegaconf.DictConfig=None,
        feedforward:    omegaconf.DictConfig=None,
        monads:         omegaconf.DictConfig=None,
        supervision:    omegaconf.DictConfig=None,
        validation:     omegaconf.DictConfig=None,
        visualization:  omegaconf.DictConfig=None,
        export:         omegaconf.DictConfig=None,
    ):
        super(Null, self).__init__(
            feedforward=feedforward, monads=monads, 
            visualization=visualization, data=data,
            supervision=supervision, validation=validation, 
            export=export, parameters=parameters,
        )
    
    def forward(self, 
        td: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        return td
        


import torch
import typing

__all__ = ['Adai']

class Adai(torch.optim.Optimizer):
    r"""Implements Adaptive Inertia Estimation (Adai) algorithm.
    
    - **Paper**: [Adai: Separating the Effects of Adaptive Learning Rate and Momentum Inertia](https://arxiv.org/pdf/2006.15815.pdf)
    - **Implementation**: [GitHub @ zeke-xie](https://github.com/zeke-xie/adaptive-inertia-adai)

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float): learning rate
        betas (Tuple[float, float], optional): beta0 and beta2 (default: (0.1, 0.99))
        eps (float, optional): the inertia bound (default: 1e-03)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)

    """

    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter], 
        lr=None,#=torch.optim.optimizer.required, 
        betas: typing.Tuple[float, float]=(0.1, 0.99), 
        eps: float=1e-03,
        weight_decay: float=0,
    ):
        #if lr is not torch.optim.optimizer.required and lr < 0.0:
        if lr is not None and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0]:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super(Adai, self).__init__(params, defaults)
    

    def __setstate__(self, state):
        super(Adai, self).__setstate__(state)
            
    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        param_size = 0
        exp_avg_sq_hat_sum = 0.
        
        for group in self.param_groups:
        
            for p in group['params']:
                if p.grad is None:
                    continue
                param_size += p.numel()
                grad = p.grad.data
                
                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p.data, memory_format=torch.preserve_format)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p.data, memory_format=torch.preserve_format)
                    # Exponential moving average of squared gradient values with the bias correction
                    state['exp_avg_sq_hat'] = torch.zeros_like(p.data, memory_format=torch.preserve_format)
                    # Cumulative products of beta1
                    state['beta1_prod'] = torch.ones_like(p.data, memory_format=torch.preserve_format)
                    

                exp_avg_sq = state['exp_avg_sq']
                exp_avg_sq_hat = state['exp_avg_sq_hat']
                beta0, beta2 = group['betas']

                state['step'] += 1
                bias_correction2 = 1 - beta2 ** state['step']

                if group['weight_decay'] != 0:
                    grad.add_(group['weight_decay'], p.data)
                    
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                
                exp_avg_sq_hat.mul_(0.).add_(exp_avg_sq / bias_correction2)
                
                exp_avg_sq_hat_sum += exp_avg_sq_hat.sum()
                
        exp_avg_sq_hat_mean = exp_avg_sq_hat_sum / param_size
        
        for group in self.param_groups:
        
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                
                state = self.state[p]

                exp_avg = state['exp_avg']
                exp_avg_sq_hat = state['exp_avg_sq_hat']
                beta1_prod = state['beta1_prod']
                beta0, beta2 = group['betas']
                beta1 = (1. - (exp_avg_sq_hat / exp_avg_sq_hat_mean).mul(beta0)).clamp(0., 1 - group['eps'])
                
                beta1_prod.mul_(beta1)
                bias_correction1 = 1 - beta1_prod
                
                exp_avg.mul_(beta1).addcmul_(1 - beta1, grad)
                exp_avg_hat = exp_avg / bias_correction1
                
                step_size = group['lr'] 
                p.data.add_(-step_size, exp_avg_hat)

        return loss


from moai.validation.collection import Metrics as Collection
from moai.validation.indicators import Indicators
from moai.validation.single import Metric as Single
from moai.validation.noop import NoOp
from moai.validation.classification import Classification

__all__ = [
    "Collection",
    "Indicators",
    "Single",
    "NoOp",
    "Classification",
]


import torch

__all__ = ["NormalizedPositionError"]

def _normalised_position_error(
    gt: torch.Tensor,
    pred: torch.Tensor,
):
    l2_norm = torch.linalg.norm(gt - pred, ord=2, dim=-1)
    return l2_norm / (torch.linalg.norm(gt, ord=2, dim=-1) + 1e-7)

#TODO: extract eps to param
class NormalizedPositionError(torch.nn.Module):
    def __init__(self,

    ):
        super(NormalizedPositionError, self).__init__()

    def forward(self,     
        gt: torch.Tensor,
        pred: torch.Tensor,
    ) -> torch.Tensor:
        return _normalised_position_error(gt, pred).mean()


from moai.utils.arguments import (
    ensure_path,
    ensure_choices,
)

import numpy as np
import torch
import typing
import logging
import os
import toolz

__all__ = ["Npz"]

log = logging.getLogger(__name__)

class Npz(typing.Callable[[typing.Dict[str, typing.Union[torch.Tensor, typing.Dict[str, torch.Tensor]]]], None]):

    __MODES__ = ['all', 'append']
    
    def __init__(self,
        path:           str,
        keys:           typing.Union[str, typing.Sequence[str]],
        mode:           str="combined", # all"
        counter_format: str="05d",
        combined:       bool=False,
        compressed:     bool=False,
    ):
        self.mode = ensure_choices(log, "saving mode", mode, Npz.__MODES__)
        self.folder = ensure_path(log, "output folder", path)
        self.index = 0
        self.keys = [keys] if type(keys) is str else list(keys)
        self.fmt = counter_format
        self.dict_mode = combined
        self.compressed = compressed
        
    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        arrays = { }
        for key in self.keys:
            split = key.split('.')
            arrays[key] = toolz.get_in(split, tensors).squeeze().detach().cpu().numpy()
        if self.mode == 'all':
            mode = 'ab'
            if self.dict_mode:
                f = os.path.join(self.folder, f"{self.index:{self.fmt}}.npz")
                np.savez_compressed(f, **arrays) if self.compressed else np.savez(f, **arrays)
            else:
                for key in self.keys:
                    f = os.path.join(self.folder, f"{self.index:{self.fmt}}_{key}.npz")
                    np.savez_compressed(f, arrays[key]) if self.compressed else np.savez(f, arrays[key])
            self.index += 1
        else:
            log.error('Npz exporting is not yet enabled in append mode.')
        


import torch
import typing

class Stack(torch.nn.Module):
    def __init__(self,
        dim: int=0
    ):
        super(Stack, self).__init__()
        self.dim = dim

    def forward(self, tensors: typing.List[torch.Tensor]) -> torch.Tensor:
        return torch.stack(tensors, dim=self.dim)

class Concat(torch.nn.Module):
    def __init__(self,
        dim: int=1
    ):
        super(Concat, self).__init__()
        self.dim = dim

    def forward(self, tensors: typing.List[torch.Tensor]) -> torch.Tensor:
        return torch.cat(tensors, dim=self.dim)

class Split(torch.nn.Module): #TODO: optimize by returning the tuple
    def __init__(self,
        dim: int=0, # at which dim to split
        split: int=0, # how many to split to, with 0 denoting an even split
    ):
        super(Split, self).__init__()
        self.dim = dim
        self.split = split

    def forward(self, 
        tensor: torch.Tensor,
        index: torch.Tensor # scalar tensor denoting the split index
    ) -> torch.Tensor:
    # ) -> typing.Tuple[torch.Tensor, torch.Tensor]:
        size = self.split if self.split else tensor.shape[self.dim] // 2
        ret = torch.split(tensor, size, dim=self.dim)[int(index)]
        return ret.squeeze(self.dim) if ret.shape[self.dim] == 1 else ret

class SelectTensor(torch.nn.Module):
    def __init__(self):

        super(SelectTensor,self).__init__()
    
    def forward(self,
                tensors: typing.List[torch.Tensor],
                ref_tensor: torch.Tensor) -> torch.Tensor:

                for ten in tensors:
                    if ten.shape[-2:] == ref_tensor.shape[-2:]:
                        out_tensor = ten
                        break
                    else:
                       out_tensor = None
                
                return out_tensor
class Detach(torch.nn.Module):
    def __init__(self):
        super(Detach, self).__init__()

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor.detach()


import math
import torch
import typing

__all__ = ['AdaMod']

#NOTE: from https://github.com/jettify/pytorch-optimizer/blob/master/torch_optimizer/adamod.py

class AdaMod(torch.optim.Optimizer):
    r"""Implements AdaMod algorithm.

    - **Paper**: [Adaptive and Momental Bounds for Adaptive Learning Rate Methods](https://arxiv.org/pdf/1910.12249.pdf)
    - **Implementation**: [GitHub @ jettify](https://github.com/jettify/pytorch-optimizer)    

    Arguments:
        params: iterable of parameters to optimize or dicts defining
            parameter groups
        lr: learning rate (default: 1e-3)
        betas: coefficients used for computing running averages of gradient
            and its square (default: (0.9, 0.999))
        beta3: smoothing coefficient for adaptive learning rates
            (default: 0.9999)
        eps: term added to the denominator to improve numerical stability
            (default: 1e-8)
        weight_decay: weight decay (L2 penalty) (default: 0)

    Example:
        >>> import torch_optimizer as optim
        >>> optimizer = optim.AdaMod(model.parameters(), lr=0.1)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    __ https://arxiv.org/abs/1910.12249

    Note:
        Reference code: https://github.com/lancopku/AdaMod
    """

    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        beta3: float=0.999,
        eps: float=1e-8,
        weight_decay: float=0,
    ) -> None:
        if lr <= 0.0:
            raise ValueError('Invalid learning rate: {}'.format(lr))
        if eps < 0.0:
            raise ValueError('Invalid epsilon value: {}'.format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(
                'Invalid beta parameter at index 0: {}'.format(betas[0])
            )
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(
                'Invalid beta parameter at index 1: {}'.format(betas[1])
            )
        if not 0.0 <= beta3 < 1.0:
            raise ValueError('Invalid beta3 parameter: {}'.format(beta3))
        if weight_decay < 0.0:
            raise ValueError(
                'Invalid weight_decay value: {}'.format(weight_decay)
            )
        defaults = dict(
            lr=lr, betas=betas, beta3=beta3, eps=eps, weight_decay=weight_decay
        )
        super(AdaMod, self).__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        """Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    msg = 'AdaMod does not support sparse gradients'
                    raise RuntimeError(msg)

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p)
                    # Exponential moving average of actual learning rates
                    state['exp_avg_lr'] = torch.zeros_like(p)

                exp_avg, exp_avg_sq, exp_avg_lr = (
                    state['exp_avg'],
                    state['exp_avg_sq'],
                    state['exp_avg_lr'],
                )
                beta1, beta2 = group['betas']

                state['step'] += 1

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

                denom = exp_avg_sq.sqrt().add_(group['eps'])

                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                step_size = (
                    group['lr']
                    * math.sqrt(bias_correction2)
                    / bias_correction1
                )

                if group['weight_decay'] != 0:
                    p.data.add_(
                        p.data, alpha=-group['weight_decay'] * group['lr']
                    )

                # Applies momental bounds on actual learning rates
                step_size = torch.full_like(denom, step_size)
                step_size.div_(denom)
                exp_avg_lr.mul_(group['beta3']).add_(
                    step_size, alpha=1 - group['beta3']
                )
                step_size = torch.min(step_size, exp_avg_lr)
                step_size.mul_(exp_avg)

                p.data.add_(-step_size)

        return loss


import moai.nn.utils as miu
import moai.nn.activation as mia

import torch

import typing
import functools
import logging

log = logging.getLogger(__name__)

__all__ = [
    "make_conv_op",
    "make_conv_1x1",
    "make_conv_3x3",
    "make_conv_block",
]

__CONV_FACTORY__ = {
    "conv2d":           torch.nn.Conv2d,
}

def _update_conv_op(name: str, type: typing.Type):
    if name not in __CONV_FACTORY__.keys():
        __CONV_FACTORY__.update({name: type})
    else:
        log.error(f"Trying to add an already existing key {name} in the convolution operation factory.")

def make_conv_op(
        convolution_type: str,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int=1,
        padding: int=1,
        dilation: int=1,        
        groups: int=1, 
        bias: bool=True,
        **kwargs
    ) -> torch.nn.Module:
        if convolution_type in __CONV_FACTORY__.keys():
            return miu.instantiate(__CONV_FACTORY__[convolution_type], 
            **{
                **locals(),
                 **kwargs
            })
        else:
            log.error(f"Convolution type {convolution_type} not found.")

make_conv_1x1 = functools.partial(make_conv_op, #NOTE: only pass conv_type and inout features
    kernel_size=1,
    stride=1,
    dilation=1,
    padding=0,
    groups=1,
    bias=False
)

make_conv_3x3 = functools.partial(make_conv_op, #NOTE: only pass conv_type and inout features
    kernel_size=3,
    stride=1,
    dilation=1,
    padding=1,
    groups=1,
    bias=True
)

__CONV_BLOCK_FACTORY__ = {

}

def _update_conv_block(name: str, type: typing.Type):
    if name not in __CONV_BLOCK_FACTORY__.keys():
        __CONV_BLOCK_FACTORY__.update({name: type})
    else:
        log.error(f"Trying to add an already existing key {name} in the convolution block factory.")

def make_conv_block(
    block_type: str,
    convolution_type: str,
    in_features: int,
    out_features: int,
    activation_type: str,
    convolution_params: dict={"kernel_size": 3},
    activation_params: dict={"inplace": True},
    **kwargs
) -> torch.nn.Module:
    if block_type in __CONV_BLOCK_FACTORY__.keys():
        return miu.instantiate(__CONV_BLOCK_FACTORY__[block_type],
            **{ 
                **locals(),
                "convolution_params": convolution_params, #TODO: merge /w kwargs?
                "activation_params": activation_params,
                **kwargs #TODO: merge to conv_params if otherwise unused?
            })
    else:
        log.error(f"Convolutional block type {block_type} not found.")

import moai.nn.convolution.torch as mit

if "conv2d" not in __CONV_BLOCK_FACTORY__.keys(): # CONV OP
    _update_conv_block("conv2d", mit.Conv2dBlock)

del mit

#TODO: "coord_conv":
#TODO: "partial_conv":

import moai.nn.convolution.spherical as mis

if "sconv2d" not in __CONV_FACTORY__.keys():
    _update_conv_op("sconv2d", mis.SphericalConv2d)

#TODO: "coord_conv":
#TODO: "partial_conv":


import torch

#NOTE: from https://arxiv.org/pdf/1703.06868.pdf

__all__ = [
    "AdaIN",
]

#TODO: To be tested
class AdaIN(torch.nn.Module):
    def __init__(self,
        epsilon: float=1e-5, #avoid division by zero.
    ):
        super(AdaIN, self).__init__()
        self.epsilon = epsilon

    def calc_vector_mean_std(self, x):
        std = torch.sqrt(torch.var(x, dim=1) + self.epsilon)
        mean = torch.mean(x, dim=1)
        return mean, std

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        size = x.size()
        x_mean, x_std = self.calc_vector_mean_std(x)
        y_mean, y_std = self.calc_vector_mean_std(y)

        normalized = (x - x_mean.unsqueeze(-1).expand(
            size)) / x_std.unsqueeze(-1).expand(size)
        return normalized * y_std.unsqueeze(-1).expand(size) + y_mean.unsqueeze(-1).expand(size)



import collections
import torch
import typing

__all__ = ['AdaShift']

#NOTE: modified from https://github.com/MichaelKonobeev/adashift

class AdaShift(torch.optim.Optimizer):
  r""" Implements AdaShift algorithm.

  - **Paper**: [ADASHIFT: DECORRELATION AND CONVERGENCE OF ADAPTIVE LEARNING RATE METHODS](https://arxiv.org/pdf/1810.00143.pdf)
  - **Implementation**: [GitHub @ MichaelKonobeev](https://github.com/MichaelKonobeev/adashift)

  See https://openreview.net/forum?id=HkgTkhRcKQ

  Arguments:
    params (iterable): iterable of parameters to optimize or dicts defining
      parameter groups
    lr (float, optional): learning rate
    keep_num (int): number of gradients used to compute first moment
      estimation.
    betas (Tuple[float, float], optional): coefficients used for computing
      running averages of gradient and its square
    eps (float, optional): term added to the denominator to improve
      numerical stability
    reduce_func (callable or None): function applied to squared gradients
      to further reduce the correlation. If None, no function is applied.
  """
  def __init__(self,
    params: typing.Iterator[torch.nn.Parameter],
    lr: float=1e-3,
    keep_num: int=10,
    betas: typing.Tuple[float, float]=(0.9, 0.999),
    eps: float=1e-10,
    reduce_func: typing.Callable=torch.max
  ):
    beta1, _ = betas
    exp_weight_sum = sum(beta1 ** i for i in range(keep_num))
    first_grad_weight = beta1 ** (keep_num - 1) / exp_weight_sum
    last_grad_weight = 1. / exp_weight_sum
    defaults = dict(lr=lr, keep_num=keep_num, betas=betas,
                    eps=eps, reduce_func=reduce_func,
                    first_grad_weight=first_grad_weight,
                    last_grad_weight=last_grad_weight)
    super(AdaShift, self).__init__(params, defaults)

  def step(self, closure=None):
    loss = None
    if closure is not None:
      loss = closure()

    for group in self.param_groups:
      for p in group["params"]:
        if p.grad is None:
          continue
        grad = p.grad.data
        if grad.is_sparse:
          raise RuntimeError("AdaShift does not support sparse gradients.")

        state = self.state[p]
        if len(state) == 0:
          state["step"] = 1
          state["grad_deque"] = collections.deque([grad.clone()], maxlen=group["keep_num"])
          state["exp_avg"] = torch.zeros_like(p.data)
          state["exp_avg_sq"] = torch.zeros_like(p.data)
          continue

        grad_deque = state["grad_deque"]
        exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]

        state["step"] += 1
        grad_apply = len(grad_deque) == group["keep_num"]
        offset_grad = grad_deque[0]
        grad_deque.append(grad.clone())
        if not grad_apply:
          continue

        first_grad_weight = group["first_grad_weight"]
        last_grad_weight = group["last_grad_weight"]
        beta1, beta2 = group["betas"]
        (exp_avg.sub_(first_grad_weight, offset_grad).mul_(beta1)
         .add_(last_grad_weight, grad))

        reduce_func = group["reduce_func"] or (lambda x: x)
        reduced_grad_sq = reduce_func(offset_grad.mul_(offset_grad))
        exp_avg_sq.mul_(beta2).add_(1 - beta2, reduced_grad_sq)
        bias_correction = 1 - beta2 ** (state["step"] - group["keep_num"])
        denom = exp_avg_sq.div(bias_correction).sqrt_().add_(group["eps"])

        p.data.addcdiv_(-group["lr"], exp_avg, denom)
    return loss



import moai.nn.utils as miu
import moai.nn.activation as mia

import torch
import typing
import functools
import logging

log = logging.getLogger(__name__)

__all__ = [
    "make_deconv_op",    
    "make_deconv_4x4",
    "make_deconv_block",
]

__DECONV_FACTORY__ = {
    "deconv2d":             torch.nn.ConvTranspose2d,
}

#NOTE: https://stackoverflow.com/questions/50683039/conv2d-transpose-output-shape-using-formula

def _update_deconv_op(name: str, type: typing.Type):
    if name not in __DECONV_FACTORY__.keys():
        __DECONV_FACTORY__.update({name: type})
    else:
        log.error(f"Trying to add an already existing key {name} in the deconvolution operation factory.")

def make_deconv_op(
        deconvolution_type: str,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int=1,
        padding: int=1,
        dilation: int=1,        
        groups: int=1, 
        bias: bool=True,
        **kwargs
    ) -> torch.nn.Module:
        if deconvolution_type in __DECONV_FACTORY__.keys():
            return miu.instantiate(__DECONV_FACTORY__[deconvolution_type], 
            **{
                **locals(),
                **kwargs
            })
        else:
            log.error(f"Deconvolution type ({deconvolution_type}) not found.")

make_deconv_4x4 = functools.partial(make_deconv_op, #NOTE: only pass deconv_type and inout features
    kernel_size=4,
    stride=2,
    dilation=1,
    padding=1,
    groups=1,
    bias=False
)

__DECONV_BLOCK_FACTORY__ = {

}

def _update_deconv_block(name: str, type: typing.Type):
    if name not in __DECONV_BLOCK_FACTORY__.keys():
        __DECONV_BLOCK_FACTORY__.update({name: type})
    else:
        log.error(f"Trying to add an already existing key {name} in the deconvolution block factory.")

def make_deconv_block(
    block_type: str,
    deconvolution_type: str,
    in_features: int,
    out_features: int,
    activation_type: str,
    deconvolution_params: dict={"kernel_size": 4},
    activation_params: dict={"inplace": True},
    **kwargs
) -> torch.nn.Module:
    if block_type in __DECONV_BLOCK_FACTORY__.keys():
        return miu.instantiate(__DECONV_BLOCK_FACTORY__[block_type],
            **{ 
                **locals(),
                "deconvolution_params": deconvolution_params,
                "activation_params": activation_params,
                **kwargs
            })
    else:
        log.error(f"Deconvolutional block type ({block_type}) not found.")

import moai.nn.deconvolution.torch as mit

if "deconv2d" not in __DECONV_BLOCK_FACTORY__.keys(): #NOTE: DECONV OP
    _update_deconv_block("deconv2d", mit.Deconv2dBlock)

del mit

#NOTE: "coord_conv":
#NOTE: "partial_conv":

#NOTE: "coord_conv":
#NOTE: "partial_conv":


from moai.utils.arguments import assert_choices

import torch
import logging

log = logging.getLogger(__name__)

__all__ = ["Finite"]

class Finite(torch.nn.Module):
    __CAST_OPS__ = {
        "float": lambda t: t.float(),
        "byte": lambda t: t.byte(),
        "bool": lambda t: t.bool(),
    }

    __MODES__ = ['nan', 'inf', 'nan+inf', 'finite']

    def __init__(self,
        mode:   str='finite', # one of ['nan', 'inf', 'nan+inf', 'finite']        
        dtype:  str="float", # "byte" #TODO: change to 'mask' and 'weights' semantics?
    ):
        super(Finite, self).__init__()
        assert_choices(log, "mode", mode, Finite.__MODES__)
        assert_choices(log, "cast type", dtype, Finite.__CAST_OPS__.keys())
        self.check_func = torch.isnan if mode == 'nan' else (
            torch.isinf if mode == 'inf' else(
                torch.isfinite if mode == 'finite' else\
                   lambda x: torch.isnan(x) & torch.isinf(x)
            )
        )
        self.cast_op = Finite.__CAST_OPS__[dtype]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.cast_op(self.check_func(x).all(dim=1, keepdim=True))
        


from moai.utils.arguments import assert_choices

import torch
import logging

log = logging.getLogger(__name__)

class Passthrough(torch.nn.Module):

    __MODES__ = ['minimize', 'maximize']

    def __init__(self, 
        mode:               str='minimize', # one of ['minimize', 'maximize']
    ):
        super(Passthrough, self).__init__()
        assert_choices(log, "mode", mode, Passthrough.__MODES__)
        self.fwd_func = lambda t: t if mode == 'minimize'\
            else lambda t: -t

    def forward(self, error: torch.Tensor) -> torch.Tensor:
        return self.fwd_func(error)



import torch
import typing

__all__ = ['ModelParameterSelector']

class ModelParameterSelector(typing.Callable[[torch.nn.Module], typing.List[torch.Tensor]]):
    def __init__(self,
        keys:       typing.Sequence[str],
    ):
        self.keys = keys        

    def __call__(self, module: torch.nn.Module) -> typing.List[torch.Tensor]:
        params = []
        for key in self.keys:
            t = module.predictions[key]
            t.requires_grad_(True)
            params.append(t)
        return params


from pytorch_lightning.callbacks import Callback
import moai.utils.engine as mieng

import omegaconf.omegaconf
import logging

log = logging.getLogger(__name__)

__all__ = ["Visualizers"]

class Visualizers(mieng.Collection, mieng.Interval):
    def __init__(self,
        batch_interval:int,
        visualizers: omegaconf.DictConfig,
    ):
        mieng.Interval.__init__(self, batch_interval)
        mieng.Collection.__init__(
            self, 
            items=visualizers, 
            name="visualizers"
        )


import moai.nn.convolution as mic
import moai.nn.residual as mires
import moai.nn.utils as miu
import moai.nn.sampling.spatial.downsample as mids
import moai.nn.sampling.spatial.upsample as mius

import torch

#NOTE: from https://github.com/anibali/pytorch-stacked-hourglass/blob/master/src/stacked_hourglass/model.py
#NOTE: from https://github.com/princeton-vl/pytorch_stacked_hourglass/blob/master/models/posenet.py

__all__ = ["Hourglass"]

class Hourglass(torch.nn.Module):
    def __init__(self, 
        convolution:        str="conv2d",
        activation:         str="relu_bn2d",
        downscale:          str="maxpool2d",
        upscale:            str="upsample2d",
        residual:           str="preactiv_bottleneck",
        depth:              int=4,
        features:           int=128,
    ):
        super(Hourglass, self).__init__()
        new_features = features
        self.up1 = mires.make_residual_block(
            block_type=residual,
            convolution_type=convolution,
            in_features=features,
            out_features=features,
            bottleneck_features=features,
            activation_type=activation,
            strided=False,
        )
        # Lower branch        
        self.pool1 = mids.make_downsample(
            downscale_type=downscale,
            features=features,
            kernel_size=3 if downscale == 'maxpool2d_aa' else 2,
        )
        self.low1 = mires.make_residual_block(
            block_type=residual,
            convolution_type=convolution,
            in_features=features,
            out_features=new_features,
            bottleneck_features=new_features,
            activation_type=activation,
            strided=False,
        )
        self.depth = depth
        # Recursive hourglass        
        self.low2 = Hourglass(
                depth=depth-1, 
                features=new_features,
                upscale = upscale,#added
                convolution=convolution,
                activation=activation,
                downscale=downscale,
                residual=residual,        
            ) if self.depth > 1 \
            else mires.make_residual_block(
                block_type=residual,
                convolution_type=convolution,
                in_features=new_features,
                out_features=new_features,
                bottleneck_features=new_features,
                activation_type=activation,
                strided=False,
            )
        self.low3 = mires.make_residual_block(
                block_type=residual,
                convolution_type=convolution,
                in_features=new_features,
                out_features=features,
                bottleneck_features=features,
                activation_type=activation,
                strided=False,
            )
        self.up2 = mius.make_upsample(
            upscale_type=upscale,
            features=features,
            mode="bilinear" #NOTE or nearest? or extract as config param?
        )

    def forward(self, x):
        up1  = self.up1(x)
        pool1 = self.pool1(x)
        low1 = self.low1(pool1)
        low2 = self.low2(low1)
        low3 = self.low3(low2)
        up2  = self.up2(low3)
        return up1 + up2


from moai.monads.generation.grid import Grid

import torch
import numpy as np

class Spherical(Grid):
    __GRID_MAP__ = {
        'pi':               'ndc',
        'tau':              'norm', 
    }

    __ORDER_MAP__ = {
        'longlat':       'xy',
        'latlong':       'yx',
    }

    __SCALE_MAP__ = {
        'longlat': {
            'pi':       [np.pi, 0.5 * np.pi],
            'tau':      [2.0 * np.pi, np.pi],
        },
        'latlong': {
            'pi':       [0.5 * np.pi, np.pi],
            'tau':      [np.pi, 2.0 * np.pi],
        },
    }

    def __init__(self,
        mode:               str='pi', # one of ['pi', 'tau']
        width:              int=512,
        inclusive:          bool=False,
        order:              str='longlat', # one of ['longlat', 'latlong']
        long_offset_pi:     float=0.0, # offset for longitude, relative to pi
    ):
        super(Spherical, self).__init__(
            mode=Spherical.__GRID_MAP__[mode],
            width=width, height=width // 2, depth=1,
            inclusive=inclusive,
            order=Spherical.__ORDER_MAP__[order],
        )
        scale = torch.Tensor([Spherical.__SCALE_MAP__[order][mode]])
        self.grid = self.grid * scale.unsqueeze(-1).unsqueeze(-1)
        if long_offset_pi != 0.0:
            dim = 0 if order.startswith('long') else 1
            self.grid[:, dim, ...] += np.pi * long_offset_pi



import torch
import typing
import logging
import toolz
from contextlib import ContextDecorator

__all__ = ['ModuleMode']

log = logging.getLogger(__name__)

class nograd_mode(torch.no_grad):
    def __init__(self, *args, **kwargs):
        super(nograd_mode, self).__init__()

class inference_mode(torch.inference_mode):
    def __init__(self, *args, **kwargs):
        super(inference_mode, self).__init__()

class eval_mode(ContextDecorator):
    def __init__(self, module: torch.nn.Module):
        self.module = module
        # self.module_train_state = False

    def __enter__(self):
        if self.module.training:
            # self.module_train_state = self.module.training
            self.module.eval()
            log.info(f'{type(self.module)} set to eval() mode.')

    def __exit__(self,
        exc_type: typing.Any,
        exc_value: typing.Any,
        traceback: typing.Any
    ) -> None:
        pass
        # if self.module_train_state:
        #     self.module.train(self.module_train_state)

class eval_nograd_mode(ContextDecorator):
    def __init__(self, module: torch.nn.Module):
        # super(eval_nograd_mode, self).__init__()
        self.module = module
        self.module_grad_state = False

    def __enter__(self):
        # self.module_train_state = self.module.training
        self.module_grad_state = torch.is_grad_enabled()
        torch.set_grad_enabled(False)
        if self.module.training:
            self.module.eval()
            log.info(f'{type(self.module)} set to eval() mode.')
        
    def __exit__(self,
        exc_type: typing.Any,
        exc_value: typing.Any,
        traceback: typing.Any
    ) -> None:
        torch.set_grad_enabled(self.module_grad_state)

class ModuleMode(typing.Callable[[torch.nn.Module], None]):
    
    __TYPE__ = {
        'nograd':              nograd_mode,
        'no_grad':             nograd_mode,
        'inference':           inference_mode,
        'infer':               inference_mode,
        'eval':                eval_mode,
        'eval_nograd':         eval_nograd_mode,
        'nograd_eval':         eval_nograd_mode,
    }

    def __init__(self,
        modules:    typing.Mapping[str, str],
    ):
        self.modules = modules

    def __call__(self, model: torch.nn.Module) -> None:
        for module, mode in self.modules.items():
            split = module.split('.')
            m = toolz.reduce(getattr, split, model)
            m.forward = ModuleMode.__TYPE__[mode](m)(m.forward)
            log.info(f"Transforming module {module}'s forward to {mode}")



import itertools
import typing
import toolz

__all__ = ["repeat"]

def repeat(
    n: int,
    value: typing.Union[typing.Any, typing.Sequence[typing.Any]]
) -> typing.Sequence[typing.Any]:
    return list(
        itertools.repeat(value, n)\
            if isinstance(value, list) else toolz.take(n, value)
    )



import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import inspect
import typing
import itertools
import toolz
import logging

log = logging.getLogger(__name__)

__all__ = ['Metrics']

from moai.monads.execution.cascade import _create_accessor

class Metrics(torch.nn.ModuleDict):
    execs: typing.List[typing.Callable] = []

    def __init__(self, 
        metrics: omegaconf.DictConfig={},
        **kwargs: typing.Mapping[str, typing.Any]
    ):
        super(Metrics, self).__init__()
        if not len(metrics):
            log.warning("A collection of metrics is being used for validating the model, but no metrics have been assigned")
        loop = ((key, params) for key, params in kwargs.items() if key in metrics)
        #TODO: add message in case key is not found
        for k, p in loop:
            self.add_module(k, hyu.instantiate(getattr(metrics, k)))
            # last_module = toolz.last(self.modules()) # moduledict is ordered            
            last_module = self[k]
            sig = inspect.signature(last_module.forward)
            if 'out' not in p:
                p['out'] = [k]
            for keys in zip(*toolz.remove(lambda x: not x, list(p[prop] for prop in itertools.chain(sig.parameters, ['out']) if p.get(prop) is not None))):
                accessors = [_create_accessor(k if isinstance(k, str) else toolz.get(0, k, None)) for k in keys[:-1]]
                self.execs.append(lambda tensor_dict, metric_dict,
                    acc=accessors, k=keys, p=sig.parameters.keys(), f=last_module:
                    metric_dict.update({
                        k[-1]: f(**dict(zip(p, 
                            list(
                                a(tensor_dict) for a, i in zip(acc, k[:-1])
                                if i is not None or None
                            )
                        )))
                    })
                )

    def forward(self,
        tensors: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        metrics = { }
        for exe in self.execs:
            exe(tensors, metrics)
        returned = { }
        for k, m in metrics.items():
            returned[f'{k}'] = m
        return returned


import functools
import math
import torch
import typing

__all__ = ['AdamSRT', 'AdamS']

#NOTE: from https://github.com/ymontmarin/adamsrt/blob/master/adamsrt/optimizers/adam_srt.py

class AdamSRT(torch.optim.Optimizer):
    """Implements the AdamSRT algorithm.

    - **Paper**: [Spherical Perspective on Learning with Batch Norm. New methods : AdamSRT](https://arxiv.org/pdf/2006.13382.pdf)
    - **Implementation**: [GitHub @ ymontmarin](https://github.com/ymontmarin/adamsrt)

    General version of Adam-SRT that works for different normalization layer
    if specific channel options (channel_dims, channel_wise, channel_gloabal)
    are given.
    It should be used on parameters that are subject to scale invariance
    because they are followed by a normalization layer.
    Because not all params are concern, group_parameters of pytorch
    should be used.
    The effect is to adapt moments of Adam to the geometry implied by
    normalization layer. RT transform the order one moment ; make the
    order 2 moment rescaled and by norm.

    Example:
        >>> par_groups = [{'params': model.conv_params(), 'channel_wise'=True},
        >>>               {'params': model.other_params()}]
        >>> optimizer = AdamSRT(par_groups, lr=0.01, betas=(0.9, 0.9999))
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    Arguments:
        params (list of dict or iterator): either a list of group_parameters
            which is a dict with key 'params' with a param iterator and other
            key for optimizer parameters overiding for this group of param,
        lr (float): learning rate,
        betas (tuple(float)): momentum factor for Adam moments,
        eps (float): float to avoid numerical instality in normalization,
        weight_decay (float): value for L2 regularization,
        channel_dims (list of int): the index of shape that represent the
            distinct dim that are independently normalized. Default value is
            channel_dims=shape which correspond to classic Adam.
            It can be used to adapt Adam to any normalization layers that
            follow conv layers,
        channel_wise (bool): if True and channel_dims is None set it to [0]
            which correspond to classic channel shape in 2D conv Network.
            Normalization will be done over other dims which are subject to
            scale invariance thanks to following normalization layer,
    """
    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        eps: float=1e-8,
        weight_decay: float=0,
        channel_dims=None,  # For customize the dimension for group of param
        channel_wise: bool=True,  # For default conv followed by BN invariance
        rt: bool=True
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))

        defaults = dict(
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            channel_dims=channel_dims,
            channel_wise=channel_wise,
            rt=rt,
        )
        super(AdamSRT, self).__init__(params, defaults)

    def step(self):
        """
        Performs a single optimizatino step
        """
        for group in self.param_groups:
            for p in group['params']:

                # Get grad of params to update
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError(
                        'RMSprop does not support sparse gradients'
                    )
                if group['weight_decay'] != 0.:
                    grad.add_(p.data, alpha=group['weight_decay'])

                # Get state
                state = self.state[p]

                # State initialization if needed
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                    # Create the scalar product in respect of channel dims
                    shape = p.data.shape
                    channel_dims = group['channel_dims']
                    if channel_dims is None:
                        if group['channel_wise']:
                            # Classic meaning of channels
                            channel_dims = [0]
                        else:
                            # element wise : every element is a channel
                            # It correpond to classic adam update
                            channel_dims = list(range(len(shape)))
                    state['channel_dims'] = channel_dims
                    state['shape'] = shape

                # Start by increment step
                state['step'] += 1

                # Create the appropriate dot operator for the invar groups
                dot_ope = self.get_dot_operator(
                    state['channel_dims'], state['shape']
                )

                # Retrive moments and constant
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                b1, b2 = group['betas']

                # Update momentums
                exp_avg.mul_(b1).add_(grad, alpha=1 - b1)
                exp_avg_sq.mul_(b2).add_(dot_ope(grad, grad), alpha=1 - b2)
                # It should be d^-1 * (1 - beta2) instead
                # To avoid double div with > 1 we multiply stepsize by d^1/2

                # Compute bias correction
                bias_correction1 = 1 - b1 ** state['step']
                bias_correction2 = 1 - b2 ** state['step']

                # Compute actual nom and denom for the step
                nom = exp_avg / bias_correction1
                denom = (
                    exp_avg_sq.sqrt() / math.sqrt(bias_correction2)
                ).add_(group['eps'])

                # Prepare data temporary copy for RT transform if needed
                if dot_ope.dim > 1 and group['rt']:
                    prev_data = p.data.clone().detach()

                # Take the step on the datas
                step_size = group['lr'] * math.sqrt(dot_ope.dim)
                p.data.addcdiv_(nom, denom, value=-step_size)

                # We are on a sphere, we do RT transform
                if dot_ope.dim > 1 and group['rt']:
                    new_data = p.data
                    prev_norm_sq = dot_ope(prev_data, prev_data)
                    new_norm_sq = dot_ope(new_data, new_data)
                    scal_x1_x2 = dot_ope(prev_data, new_data)
                    scal_m_x2 = dot_ope(exp_avg, new_data)
                    # R  order 2 moment
                    (
                        exp_avg_sq
                        .mul_(prev_norm_sq)
                        .div_(new_norm_sq + group['eps'])
                    )
                    # RT the order 1 moment
                    (
                        exp_avg
                        .mul_(scal_x1_x2)
                        .add_(-scal_m_x2 * prev_data)
                        .div_(new_norm_sq + group['eps'])
                    )

    @staticmethod
    def get_dot_operator(channel_dims, shape):
        """
        Generate a function that do scalar product for each channel dims
        Over the remaining dims
        """
        # Other dims are the ones of groups of elem for each channel
        grp_dims = list(set(range(len(shape))) - set(channel_dims))

        # Compute shape and size
        channel_shape = [shape[i] for i in channel_dims]
        grp_shape = [shape[i] for i in grp_dims]
        channel_size = functools.reduce(lambda x, y: x * y, [1] + channel_shape)
        grp_size = functools.reduce(lambda x, y: x * y, [1] + grp_shape)

        # Prepare the permutation to ordonate dims and its reciproc
        perm = channel_dims + grp_dims
        antiperm = [
            e[1]
            for e in sorted([(j, i) for i, j in enumerate(perm)])
        ]

        # Prepare index query that retrieve all dimensions
        slice_len = max(len(channel_shape), 1)
        idx = [slice(None)] * slice_len + [None] * (len(shape) - slice_len)

        # Define the scalar product channel wise over grp dims
        # Output have is extend to fit initial shape
        def scalar_product(tensor1, tensor2):
            return (
                (tensor1 * tensor2)
                .permute(perm)  # permute as chan_dims, grp_dims
                .contiguous()
                .view(channel_size, grp_size)  # view as 2 dims tensor
                .sum(dim=1)  # norm over group dims to have scalar
                .view(*(channel_shape if channel_shape else [-1]))
                [idx]  # restore channel shape and extend on grp dims
                .permute(antiperm)  # Reverse permute to retrieve shape
                .contiguous()
            )
        scalar_product.dim = grp_size
        return scalar_product

class AdamS(AdamSRT):
    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        eps: float=1e-8,
        weight_decay: float=0,
        channel_dims=None,  # For customize the dimension for group of param
        channel_wise: bool=True,  # For default conv followed by BN invariance
    ):
        super(AdamS, self).__init__(
            params,
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            channel_dims=channel_dims,
            channel_wise=channel_wise,
            rt=False  # Never do RT
        )


import random
import os
import numpy as np
import torch
import logging
import pytorch_lightning as pl

log = logging.getLogger(__name__)

#NOTE: modified from https://github.com/wbaek/torchskeleton/blob/master/skeleton/utils/utils.py
#       and https://github.com/PyTorchLightning/PyTorch-Lightning/blob/0.7.6/pytorch_lightning/trainer/seed.py#L32 

__all__ = ["ManualSeed", "RandomSeed"]

def _set_random_seed_all(seed: int, deterministic: bool=False) -> None:    
    os.environ['PYTHONHASHSEED'] = str(seed)
    os.environ["PL_GLOBAL_SEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    #pl.seed_everything(seed)
    if deterministic:
        # os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:8' #NOTE: see https://docs.nvidia.com/cuda/cublas/index.html#cublasApi_reproducibility 
        # torch.set_deterministic(True) #NOTE: see https://pytorch.org/docs/stable/generated/torch.set_deterministic.html#torch.set_deterministic 
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

class ManualSeed(object):
    def __init__(self, 
        seed: int,
        deterministic: bool
    ):
        log.info(f"Setting the seed to {seed}.")
        _set_random_seed_all(seed=seed, deterministic=deterministic)

class RandomSeed(object):
    def __init__(self):
        max_seed_value = np.iinfo(np.uint32).max
        min_seed_value = np.iinfo(np.uint32).min
        seed = random.randint(min_seed_value, max_seed_value)
        log.warning(f"No seed was manually specified, seed set randomly to {seed}")        
        _set_random_seed_all(seed)


from moai.monads.keypoints.tranform_coords import (
    NormToCoords, 
    NormToNdc, 
    NdcToCameraCoords,
    CoordsToNorm,
    ScaleCoords,
    UpscaleCoords_x2,
    UpscaleCoords_x4,
    DownscaleCoords_x2,
    DownscaleCoords_x4
)

from moai.monads.keypoints.quantize_coords import (
    QuantizeCoords
)

from moai.monads.keypoints.visibility import (
    VisibilityFOV,
    VisibilityHeatmap
)

__all__ = [
    "NormToCoords",
    "NormToNdc",
    "NdcToCameraCoords",
    "CoordsToNorm",
    "QuantizeCoords",
    "VisibilityFOV",
    "VisibilityHeatmap",
    "ScaleCoords",
    "UpscaleCoords_x2",
    "UpscaleCoords_x4",
    "DownscaleCoords_x2",
    "DownscaleCoords_x4"
]


from moai.monads.filter.highpass.finite_difference import FiniteDifference
from moai.monads.filter.highpass.image2d import (
    Laplacian2d,
    Sobel2d,
    Diff2d,
)

import functools

Backward1d = functools.partial(FiniteDifference, dims=[2], mode='backward')
Forward1d = functools.partial(FiniteDifference, dims=[2], mode='forward')
Central1d = functools.partial(FiniteDifference, dims=[2], mode='central')

Backward2d = functools.partial(FiniteDifference, dims=[2, 3], mode='backward')
Forward2d = functools.partial(FiniteDifference, dims=[2, 3], mode='forward')
Central2d = functools.partial(FiniteDifference, dims=[2, 3], mode='central')

Backward3d = functools.partial(FiniteDifference, dims=[2, 3, 4], mode='backward')
Forward3d = functools.partial(FiniteDifference, dims=[2, 3, 4], mode='forward')
Central3d = functools.partial(FiniteDifference, dims=[2, 3, 4], mode='central')

__all__ = [
    "FiniteDifference",
    "Backward1d",
    "Forward1d",
    "Central1d",
    "Backward2d",
    "Forward2d",
    "Central2d",
    "Backward3d",
    "Forward3d",
    "Central3d",
    "Laplacian2d", 
    "Sobel2d", 
    "Diff2d",
]


from moai.monads.utils import dim_list
from moai.utils.arguments import assert_numeric

import torch
import functools
import logging

log = logging.getLogger(__name__)

#NOTE: beta: from https://openreview.net/forum?id=Sy2fzU9gl
#NOTE: capacity: from https://arxiv.org/pdf/1804.03599.pdf
#NOTE: robust: from https://arxiv.org/pdf/2007.13886.pdf

__all__ = ["StandardNormalKL"]

class StandardNormalKL(torch.nn.Module):
    r""" Implements the generalized Kullback-Leibler (KL) divergence function assuming the standard normal distribution as prior, including its **beta** (β-VAE), **capacity** (disentangled β-VAE), and **robust** (Charbonnier) variants. The method receives the mean (μ) and the variance (σ) of the input distribution as input and returns the KL divergence.

    ??? note "Standard KL Loss"
        <p align="center"><img src="https://render.githubusercontent.com/render/math?math=%5Chuge%5Cbegin%7Bequation%7D%5Cmathrm%7BStandardKL%7D%28%5Cmu%2C%5Csigma%29%20%3D%20%5Cfrac%7B1%7D%7B2%7D%20%5Cdisplaystyle%5Csum_%7Bi%7D%20%281%2B%5Clog%28%5Csigma_%7Bi%7D^%7B2%7D%29%20-%5Cmu_%7Bi%7D^%7B2%7D-%5Csigma_%7Bi%7D^%7B2%7D%29%5Cend%7Bequation%7D"/></p>

    ??? note "Beta-Standard KL Loss"
        <p align="center"><img src="https://render.githubusercontent.com/render/math?math=%5Chuge%5Cbegin%7Bequation%7D%5Cbeta%5Cmathrm%7B-StandardKL%7D%28%5Cmu%2C%5Csigma%29%20%3D%20%5Cbeta%20%5C%2C%20%5C%2C%20%5Cfrac%7B1%7D%7B2%7D%20%5Cdisplaystyle%5Csum_%7Bi%7D%20%281%2B%5Clog%28%5Csigma_%7Bi%7D^%7B2%7D%29%20-%5Cmu_%7Bi%7D^%7B2%7D-%5Csigma_%7Bi%7D^%7B2%7D%29%5Cend%7Bequation%7D"/></p>

    ??? note "Capacity Standard KL Loss"
        <p align="center"><img src="https://render.githubusercontent.com/render/math?math=%5Chuge%5Cbegin%7Bequation%7D%5Cmathrm%7BCapacityStandardKL%7D%28%5Cmu%2C%5Csigma%29%20%3D%20%5Cbeta%20%5C%2C%20%5C%2C%20%5C%2C%20|%5C%2C%5Cfrac%7B1%7D%7B2%7D%20%5Cdisplaystyle%5Csum_%7Bi%7D%20%281%2B%5Clog%28%5Csigma_%7Bi%7D^%7B2%7D%29%20-%5Cmu_%7Bi%7D^%7B2%7D-%5Csigma_%7Bi%7D^%7B2%7D%29%5C%2C-%5C%2CC%5C%2C|%5Cend%7Bequation%7D"/></p>

    ??? note "Robust Standard KL Loss"
        <p align="center"><img src="https://render.githubusercontent.com/render/math?math=%5Chuge%5Cbegin%7Bequation%7D%5Cmathrm%7BRobustStandardKL%7D%28%5Cmu%2C%5Csigma%29%20%3D%20%5Csqrt%7B1%20%2B%20%5Cbig%28%5Cfrac%7B1%7D%7B2%7D%20%5Cdisplaystyle%5Csum_%7Bi%7D%20%281%2B%5Clog%28%5Csigma_%7Bi%7D^%7B2%7D%29%20-%5Cmu_%7Bi%7D^%7B2%7D-%5Csigma_%7Bi%7D^%7B2%7D%29%5Cbig%29^2%7D%20-%201%5Cend%7Bequation%7D"/></p>

    ??? cite "Papers"
        [![Paper](https://img.shields.io/static/v1?label=OpenReview&message=beta-VAE: Learning Basic Visual Concepts with a Constrained Variational Framework&color=1c1ca2&)](https://openreview.net/forum?id=Sy2fzU9gl)

        [![Paper](https://img.shields.io/static/v1?label=1904.07399&message=Understanding disentangling in β-VAE&color=1c1ca2&logo=arxiv)](https://arxiv.org/pdf/1804.03599.pdf)
        
        [![Paper](https://img.shields.io/static/v1?label=2006.11697&message=Perpetual Motion: Generating Unbounded Human Motion&color=1c1ca2&logo=arxiv)](https://arxiv.org/pdf/2007.13886.pdf)

    ???+ example "Configuration"
        === "Main Entry"
            ```yaml
            - model/supervision/loss/classification/distribution:  std_kl
            ```
        === "Parameters"
            ```yaml
            mode: 'standard'
            is_var_log: true
            beta: 1.0
            C_max: 0
            C_max_iter: 0      
            ```
        === "DML"
            ```yaml
            model:
              supervision:
                losses:
                  std_kl:
                    mode: capacity
                    is_var_log: true
                    beta: 100.0
                    C_max: 25
                    C_max_iter: 1e4
            ```
        === "Graph"
            ```yaml
            model:
              supervision:
                losses:
                  std_kl:
                    mu: [input_distribution_mean_tensor_name] # the name of the mean tensor
                    var: [input_distribution_variance_tensor_name] # the name of the variance tensor
                    out: [std_kl_loss] # optional, will be 'std_kl' if omitted
            ```
    
    Arguments:
        mode (str, required):
            Kullback-Leibler divergence between the input distribution and the standard normal one. Selector parameter can be one of [standard, beta, capacity, robust]. 
            See the respective papers for more details, each variant uses a different parameter subset.
            **Default value: standard.**
        is_var_log: (boolean, required):
            Incidates if the input is already in logarithmic scale (True) or not (False).
            **Default value: True**
        beta (float): 
            Represents the regularisation coefficient **β** that constrains the capacity of the latent space of a VAE model
            **Default value: 1.0**
        C_max (int):
            The maximum value of the factor capacity. Used in combination with 'capacity' mode for the disentangled VAE model
            **Default value: 0**
        C_max_iter (int):
            The iteration where the capacity value stops to increase. Used in combination with 'capacity' mode for the disentangled VAE model
            **Default value: 0**

    !!! important
        β should be set to 1.0 when using 'standard' and 'robust' modes.

    !!! important
        C_max and C_max_iter should be used only in the context of 'capacity' mode with β >> 1.0

    !!! important
        The '_robust_' variant uses the [Charbonier penalty](https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.469.129&rep=rep1&type=pdf) to prevent posterior collapse.

    
    """
    def  __init__(self,
        mode:           str = 'standard', # ['standard', 'beta', 'capacity', 'robust']
        is_var_log:     bool=True,
        beta:           float=1.0, #hyperparameter for the betaVAE case
        C_max:          int=0, #total capacity hyperparameter for the disentangled betaVAE case
        C_max_iter:     int=1e5, #hyperparameter for the disentangled betaVAE (capacity) case
    ):
        super(StandardNormalKL, self).__init__()
        assert_numeric(log, 'C_max', C_max, 0.0, None)
        assert_numeric(log, 'C_max_iter', C_max_iter, 0.0, None)
        self.is_var_log = is_var_log
        self.beta = beta
        self.C_max_iter = C_max_iter
        self.num_iter = 0
        self.register_buffer("C_max", torch.Tensor([C_max]))
        self.loss = self.loss_disentangled_beta if mode == 'capacity'\
            else (self.loss_robust_charbonnier if mode == 'robust'\
                else self.loss_beta
            )

    @staticmethod
    def loss_beta(
        mu:  torch.Tensor,
        var:        torch.Tensor,
        C_max:      torch.Tensor,
        C_max_iter: float,
        beta:       float,
        num_iter:   int
    ) -> torch.Tensor:
        normal_kl = 1.0 + var - mu.pow(2) - var.exp()
        return -0.5 * torch.sum(normal_kl, dim=dim_list(var)) * beta

    @staticmethod
    def loss_disentangled_beta(
        mu:  torch.Tensor,
        var:        torch.Tensor,
        C_max:      torch.Tensor,
        C_max_iter: float,
        beta:       float,
        num_iter:   int
    ) -> torch.Tensor:
        normal_kl = 1.0 + var - mu.pow(2) - var.exp()
        C = torch.clamp(C_max/C_max_iter * num_iter, 0, C_max.data[0])
        return torch.abs(-0.5 * torch.sum(normal_kl, dim=dim_list(var)) - C) * beta

    @staticmethod
    def loss_robust_charbonnier(
        mu:  torch.Tensor,
        var:        torch.Tensor,
        C_max:      torch.Tensor,
        C_max_iter: float,
        beta:       float,
        num_iter:   int
    ) -> torch.Tensor:
        normal_kl = 1.0 + var - mu.pow(2) - var.exp()
        s =  -0.5 * normal_kl * beta
        charbonnier_s  = torch.sqrt(1 + s.pow(2)) - 1
        return torch.sum(charbonnier_s, dim=dim_list(var))

    def forward(self,
        var: torch.Tensor,
        mu: torch.Tensor,
    ) -> torch.Tensor:
        self.num_iter += 1
        v = var if self.is_var_log else var.log()
        return self.loss(mu, v, self.C_max, self.C_max_iter, self.beta, self.num_iter)



import copy
import math
import typing
import torch

__all__ = ['A2GradUni', 'A2GradInc', 'A2GradExp']

#NOTE: from https://github.com/jettify/pytorch-optimizer/blob/master/torch_optimizer/a2grad.py

class A2GradUni(torch.optim.Optimizer):
    r"""Implements A2GradUni Optimizer Algorithm.

    - **Paper**: [Optimal Adaptive and Accelerated Stochastic Gradient Descent](https://arxiv.org/pdf/1810.00553.pdf)
    - **Implementation**: [GitHub @ jettify](https://github.com/jettify/pytorch-optimizer)

    Arguments:
        params: iterable of parameters to optimize or dicts defining
            parameter groups
        lr: not used for this optimizer (default: None)
        beta:  (default: 10)
        lips: Lipschitz constant (default: 10)


    Example:
        >>> import torch_optimizer as optim
        >>> optimizer = optim.A2GradUni(model.parameters(), lips=10)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    __ https://arxiv.org/abs/1810.00553

    Note:
        Reference code: https://github.com/severilov/A2Grad_optimizer
    """

    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: typing.Optional[float] = None,
        beta: float=10,
        lips: float=10,
    ):

        defaults = dict(beta=beta, lips=lips, lr=lr)
        # lr is not supported for this optimizer, we need to make tests work
        # and schedulers not to fail
        if beta < 0.0:
            raise ValueError('Invalid beta value: {}'.format(beta))
        if lips < 0.0:
            raise ValueError('Invalid lips value: {}'.format(lips))

        super().__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        r"""Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['alpha_k'] = 1
                    state['v_k'] = 0
                    state['avg_grad'] = copy.deepcopy(grad)
                    state['x_k'] = copy.deepcopy(p.data)

                gamma_k = 2 * group['lips'] / (state['step'] + 1)

                avg_grad = state['avg_grad']
                avg_grad.mul_(state['step'])
                avg_grad.add_(grad)
                avg_grad.div_(state['step'] + 1)

                delta_k = torch.add(grad, avg_grad, alpha=-1)

                state['v_k'] += torch.sum(delta_k * delta_k).item()

                h_k = math.sqrt(state['v_k'])
                alpha_k_1 = 2 / (state['step'] + 3)
                coef = 1 / (gamma_k + group['beta'] * h_k)
                x_k_1 = state['x_k']
                x_k_1.add_(grad, alpha=-coef)

                p.data.mul_(1 - alpha_k_1)
                p.data.add_(x_k_1, alpha=alpha_k_1)
                p.data.add_(
                    grad, alpha=-(1 - alpha_k_1) * state['alpha_k'] * coef
                )

                state['alpha_k'] = alpha_k_1
                state['step'] += 1

        return loss


class A2GradInc(torch.optim.Optimizer):
    r"""Implements A2GradInc Optimizer Algorithm.

    - **Paper**: [Optimal Adaptive and Accelerated Stochastic Gradient Descent](https://arxiv.org/pdf/1810.00553.pdf)
    - **Implementation**: [GitHub @ jettify](https://github.com/jettify/pytorch-optimizer)

    Arguments:
        params: iterable of parameters to optimize or dicts defining
            parameter groups
        lr: not used for this optimizer (default: None)
        beta:  (default: 10)
        lips: Lipschitz constant (default: 10)


    Example:
        >>> import torch_optimizer as optim
        >>> optimizer = optim.A2GradInc(model.parameters(), lips=10)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    __ https://arxiv.org/abs/1810.00553

    Note:
        Reference code: https://github.com/severilov/A2Grad_optimizer
    """

    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: typing.Optional[float]=None,
        beta: float=10,
        lips: float=10,
    ):
        if beta < 0.0:
            raise ValueError('Invalid beta value: {}'.format(beta))
        if lips < 0.0:
            raise ValueError('Invalid weight_decay value: {}'.format(lips))
        defaults = dict(beta=beta, lips=lips, lr=lr)
        super(A2GradInc, self).__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        r"""Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['alpha_k'] = 1
                    state['v_k'] = 0
                    state['avg_grad'] = copy.deepcopy(grad)
                    state['x_k'] = copy.deepcopy(p.data)

                gamma_k = 2 * group['lips'] / (state['step'] + 1)

                avg_grad = state['avg_grad']
                avg_grad.mul_(state['step'])
                avg_grad.add_(grad)
                avg_grad.div_(state['step'] + 1)

                delta_k = torch.add(grad, avg_grad, alpha=-1)

                state['v_k'] *= (state['step'] / (state['step'] + 1)) ** 2
                state['v_k'] += torch.sum(delta_k * delta_k).item()

                h_k = math.sqrt(state['v_k'])
                alpha_k_1 = 2 / (state['step'] + 3)
                coef = 1 / (gamma_k + group['beta'] * h_k)
                x_k_1 = state['x_k']
                x_k_1.add_(grad, alpha=-coef)

                p.data.mul_(1 - alpha_k_1)
                p.data.add_(x_k_1, alpha=alpha_k_1)
                p.data.add_(
                    grad, alpha=-(1 - alpha_k_1) * state['alpha_k'] * coef
                )

                state['alpha_k'] = alpha_k_1
                state['step'] += 1

        return loss


class A2GradExp(torch.optim.Optimizer):
    r"""Implements A2GradExp Optimizer Algorithm.

    - **Paper**: [Optimal Adaptive and Accelerated Stochastic Gradient Descent](https://arxiv.org/pdf/1810.00553.pdf)
    - **Implementation**: [GitHub @ jettify](https://github.com/jettify/pytorch-optimizer)

    Arguments:
        params: iterable of parameters to optimize or dicts defining
            parameter groups
        lr: not used for this optimizer (default: None)
        beta:  (default: 10)
        lips: Lipschitz constant (default: 10)
        rho: represents the degree of weighting decrease, a constant
            smoothing factor between 0 and 1 (default: 0.5)

    Example:
        >>> import torch_optimizer as optim
        >>> optimizer = optim.A2GradExp(model.parameters(), lips=10)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    __ https://arxiv.org/abs/1810.00553

    Note:
        Reference code: https://github.com/severilov/A2Grad_optimizer
    """

    def __init__(
        self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: typing.Optional[float]=None,
        beta: float=10,
        lips: float=10,
        rho: float=0.5,
    ):

        defaults = dict(beta=beta, lips=lips, rho=rho, lr=lr)
        super(A2GradExp, self).__init__(params, defaults)
        if beta < 0.0:
            raise ValueError('Invalid beta value: {}'.format(beta))
        if lips < 0.0:
            raise ValueError('Invalid lips value: {}'.format(lips))
        if rho < 0.0 or rho > 1.0:
            raise ValueError('Invalid rho value: {}'.format(rho))

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        r"""Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['alpha_k'] = 1
                    state['v_k'] = 0
                    state['avg_grad'] = copy.deepcopy(grad)
                    state['x_k'] = copy.deepcopy(p.data)

                gamma_k = 2 * group['lips'] / (state['step'] + 1)

                avg_grad = state['avg_grad']
                avg_grad.mul_(state['step'])
                avg_grad.add_(grad)
                avg_grad.div_(state['step'] + 1)

                delta_k = torch.add(grad, avg_grad, alpha=-1)

                if state['step'] == 0:
                    state['v_kk'] = torch.sum(delta_k * delta_k).item()
                else:
                    state['v_kk'] *= group['rho']
                    state['v_kk'] += (1 - group['rho']) * torch.sum(
                        delta_k * delta_k
                    ).item()
                state['v_k'] = max([state['v_kk'], state['v_k']])

                h_k = math.sqrt((state['step'] + 1) * state['v_k'])

                alpha_k_1 = 2 / (state['step'] + 3)

                coef = -1 / (gamma_k + group['beta'] * h_k)
                x_k_1 = state['x_k']
                x_k_1.add_(grad, alpha=coef)

                p.data.mul_(1 - alpha_k_1)
                p.data.add_(x_k_1, alpha=alpha_k_1)
                p.data.add_(
                    grad, alpha=(1 - alpha_k_1) * state['alpha_k'] * coef
                )

                state['alpha_k'] = alpha_k_1
                state['step'] += 1

        return loss


import torch
import logging

log = logging.getLogger(__name__)

__all__ = ['EmptyCache']

class EmptyCache(object):
    def __init__(self, 

    ):
        log.info(f"Emptying PyTorch CUDA cache.")        
        torch.cuda.empty_cache()


import torch
import functools

__all__ = [
    'BN2d',
    'ReLu_BN',
    'BN2d_ReLu',
    'ReLu_BN2d',
    'LReLu_BN',
    'BN2d_LReLu',
    'LReLu_BN2d',
    'Normalize',
]

class BN2d(torch.nn.BatchNorm2d):
    def __init__(self,
        features: int,
        momentum: float=0.1,         
        epsilon: float=1e-5,
    ):
        super(BN2d, self).__init__(
            num_features=features, eps=epsilon, momentum=momentum
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return super(BN2d, self).forward(x)

class ReLu_BN(torch.nn.Module):
    def __init__(self,
        features: int,
        inplace: bool=True,
        epsilon: float=1e-5,
    ):
        super(ReLu_BN, self).__init__()
        self.bn = torch.nn.BatchNorm1d(features, eps=epsilon)
        self.activation = torch.nn.ReLU(inplace=inplace)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.bn(self.activation(x))

class BN2d_ReLu(torch.nn.Module):
    def __init__(self,
        features: int,
        inplace: bool=True,
        epsilon: float=1e-5,
    ):
        super(BN2d_ReLu, self).__init__()
        self.bn = torch.nn.BatchNorm2d(features, eps=epsilon)
        self.activation = torch.nn.ReLU(inplace=inplace)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.activation(self.bn(x))
        
class ReLu_BN2d(torch.nn.Module):
    def __init__(self,
        features: int,
        inplace: bool=True,
        epsilon: float=1e-5,
    ):
        super(ReLu_BN2d, self).__init__()
        self.bn = torch.nn.BatchNorm2d(features, eps=epsilon)
        self.activation = torch.nn.ReLU(inplace=inplace)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.bn(self.activation(x))

class LReLu_BN(torch.nn.Module):
    def __init__(self,
        features: int,
        inplace: bool=True,
        epsilon: float=1e-5,
        negative_slope: float=0.01,
    ):
        super(LReLu_BN, self).__init__()
        self.bn = torch.nn.BatchNorm1d(features, eps=epsilon)
        self.activation = torch.nn.LeakyReLU(negative_slope=negative_slope, inplace=inplace)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.bn(self.activation(x))

class BN2d_LReLu(torch.nn.Module):
    def __init__(self,
        features: int,
        inplace: bool=True,
        epsilon: float=1e-5,
        negative_slope: float=0.01,
    ):
        super(BN2d_LReLu, self).__init__()
        self.bn = torch.nn.BatchNorm2d(features, eps=epsilon)
        self.activation = torch.nn.LeakyReLU(negative_slope=negative_slope, inplace=inplace)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.activation(self.bn(x))

class LReLu_BN2d(torch.nn.Module):
    def __init__(self,
        features: int,
        inplace: bool=True,
        epsilon: float=1e-5,
        negative_slope: float=0.01,
    ):
        super(LReLu_BN2d, self).__init__()
        self.bn = torch.nn.BatchNorm2d(features, eps=epsilon)
        self.activation = torch.nn.LeakyReLU(negative_slope=negative_slope, inplace=inplace)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.bn(self.activation(x))

class Normalize(torch.nn.Module):
    def __init__(self,
        p:          int=2,
        dim:        int=1,
        epsilon:    float=1e-12,
    ):
        super(Normalize, self).__init__()
        self.func = functools.partial(torch.nn.functional.normalize,
            p=p, dim=dim, eps=epsilon,        
        )
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.func(x)


import moai.log.lightning as milog

import pytorch_lightning
import hydra.utils as hyu
import omegaconf.omegaconf
import typing

__all__ = ["LightningTrainer"]

class LightningTrainer(pytorch_lightning.Trainer):
    def __init__(self,
        logging:                    omegaconf.DictConfig=None,
        checkpoint:                 omegaconf.DictConfig=None,
        regularization:             omegaconf.DictConfig=None,
        callbacks:                  omegaconf.DictConfig=None,
        model_callbacks:            typing.Sequence[pytorch_lightning.Callback]=None,
        default_root_dir:           typing.Optional[str]=None,
        gradient_clip_val:          float=0.0,
        gradient_clip_algorithm:    str='norm', #NOTE: @PTL1.5    
        process_position:           int=0,
        num_nodes:                  int=1,
        num_processes:              int=1,
        devices:                    typing.Optional[typing.Union[typing.List[int], str, int]]=None,
        gpus:                       typing.Optional[typing.Union[typing.List[int], str, int]]=None,
        # https://pytorch-lightning.readthedocs.io/en/latest/advanced/multi_gpu.html#select-gpu-devices
        auto_select_gpus:           bool=False,
        tpu_cores:                  typing.Optional[typing.Union[typing.List[int], str, int]]=None,
        log_gpu_memory:             typing.Optional[str]=None,
        progress_bar_refresh_rate:  int=1,
        overfit_batches:            float=0.0,
        track_grad_norm:            int=-1,
        check_val_every_n_epoch:    int=1,
        fast_dev_run:               bool=False,
        accumulate_grad_batches:    typing.Union[int, typing.Dict[int, int], typing.List[list]]=1,
        max_epochs:                 int=1000,
        min_epochs:                 int=1,
        max_steps:                  typing.Optional[int]=None,
        min_steps:                  typing.Optional[int]=None,
        limit_train_batches:        typing.Union[int, float]=1.0,
        limit_val_batches:          typing.Union[int, float]=1.0,
        limit_test_batches:         typing.Union[int, float]=1.0,
        limit_predict_batches:      typing.Union[int, float]=1.0, #NOTE: @PTL1.5
        val_check_interval:         typing.Union[int, float]=1.0,
        flush_logs_every_n_steps:   int=100,
        log_every_n_steps:          int=10,
        accelerator:                typing.Optional[typing.Union[str, pytorch_lightning.accelerators.Accelerator]]=None,
        strategy:                   str=None,
        sync_batchnorm:             bool=False,
        precision:                  int=32,
        enable_model_summary:       bool=True,
        weights_summary:            typing.Optional[str]='full',
        weights_save_path:          typing.Optional[str]=None,
        num_sanity_val_steps:       int=2,
         #NOTE: @PTL1.5 truncated_bptt_steps:       typing.Optional[int]=None,
        resume_from_checkpoint:     typing.Optional[str]=None,
        profiler:                   typing.Optional[typing.Union[pytorch_lightning.profiler.BaseProfiler, bool, str]]=None,
        benchmark:                  bool=False,
        deterministic:              bool=True,
        reload_dataloaders_every_epoch: bool=False,
        reload_dataloaders_every_n_epochs: int=0,
        auto_lr_find:               typing.Union[bool, str]=False,        
        replace_sampler_ddp:        bool=True,
        detect_anomaly:             bool=False, #NOTE: @PTL1.5
        auto_scale_batch_size:      typing.Union[str, bool]=False,
        prepare_data_per_node:      bool=True,
        plugins:                    typing.Optional[list]=None,
        amp_backend:                str='native',
        amp_level:                  str='O2',
        move_metrics_to_cpu:        bool=False, #NOTE: @PTL1.5
        terminate_on_nan:           bool=False,
        multiple_trainloader_mode:  str="max_size_cycle",
        stochastic_weight_avg:      bool=False,                        
        #NOTE: @PTL1.5 distributed_backend:        typing.Optional[str]=None,
        #NOTE: @PTL1.5 automatic_optimization:     bool=True,
        **kwargs
    ):
        if logging and '_target_' not in logging: #TODO: needs a workaround for other viz types (e.g. not visdom) if they are moved to top level
            #TODO: wrap config field setting into a helper method
            omegaconf.OmegaConf.set_struct(logging, False)
            logging['_target_'] = 'moai.log.lightning.Collection'
            omegaconf.OmegaConf.set_struct(logging, True)
        logger = hyu.instantiate(logging)\
            if logging is not None else milog.NoOp()
        pytl_callbacks = [hyu.instantiate(c) for c in callbacks.values()]\
            if callbacks is not None else []
        checkpoint_callback = False
        if checkpoint is not None:
            pytl_callbacks.append(hyu.instantiate(checkpoint))
            checkpoint_callback = True
        if regularization is not None:
            pytl_callbacks.append(hyu.instantiate(regularization))
        if model_callbacks:
            pytl_callbacks.extend(model_callbacks)
        super(LightningTrainer, self).__init__(
            logger=logger,
            checkpoint_callback=checkpoint_callback,
            callbacks=pytl_callbacks,
            default_root_dir=None if not default_root_dir else default_root_dir,
            gradient_clip_val=gradient_clip_val,
            gradient_clip_algorithm=gradient_clip_algorithm, #NOTE: @PTL1.5
            process_position=process_position,
            num_nodes=num_nodes,
            devices=devices, #NOTE @PTL1.5 #TODO: check if needed
            gpus=gpus,
            auto_select_gpus=auto_select_gpus,
            tpu_cores=tpu_cores,
            log_gpu_memory=log_gpu_memory,
            progress_bar_refresh_rate=progress_bar_refresh_rate,
            overfit_batches=overfit_batches,
            track_grad_norm=track_grad_norm,
            check_val_every_n_epoch=check_val_every_n_epoch,
            fast_dev_run=fast_dev_run,
            accumulate_grad_batches=accumulate_grad_batches,
            max_epochs=max_epochs,
            min_epochs=min_epochs,
            max_steps=max_steps,
            min_steps=min_steps,
            max_time=None, #NOTE @PTL1.5 #TODO: check if needed
            limit_train_batches=limit_train_batches,
            limit_val_batches=limit_val_batches,
            limit_test_batches=limit_test_batches,
            limit_predict_batches=limit_predict_batches, #NOTE: @PTL1.5
            val_check_interval=val_check_interval,
            flush_logs_every_n_steps=flush_logs_every_n_steps,
            log_every_n_steps=log_every_n_steps,
            accelerator=accelerator,
            strategy=strategy, #NOTE: @PTL1.5
            sync_batchnorm=sync_batchnorm,
            precision=precision,
            enable_model_summary=enable_model_summary,
            weights_summary=weights_summary,
            weights_save_path=weights_save_path,
            num_sanity_val_steps=num_sanity_val_steps,
             #NOTE: @PTL1.5 truncated_bptt_steps=truncated_bptt_steps,
            resume_from_checkpoint=resume_from_checkpoint,
            profiler=profiler,
            benchmark=benchmark,
            deterministic=deterministic,
            reload_dataloaders_every_epoch=reload_dataloaders_every_epoch,
            reload_dataloaders_every_n_epochs=reload_dataloaders_every_n_epochs,
            auto_lr_find=auto_lr_find,
            detect_anomaly=detect_anomaly, #NOTE: @PTL1.5
            replace_sampler_ddp=replace_sampler_ddp,
            terminate_on_nan=terminate_on_nan,
            auto_scale_batch_size=auto_scale_batch_size,
            prepare_data_per_node=prepare_data_per_node,
            plugins=plugins,
            amp_backend=amp_backend,            
            #NOTE: @PTL1.5 fixdistributed_backend=distributed_backend,
            amp_level=amp_level,
            #NOTE: @PTL1.5 fixautomatic_optimization=automatic_optimization,
            move_metrics_to_cpu=move_metrics_to_cpu,
            multiple_trainloader_mode=multiple_trainloader_mode,
            stochastic_weight_avg=stochastic_weight_avg,
            num_processes=num_processes, #NOTE: @PTL1.5 fix
            **kwargs
        )

    def run(self, model):
        self.fit(model)


"""
Copyright (C) 2020 Jeremy Bernstein, Arash Vahdat, Yisong Yue & Ming-Yu Liu.  All rights reserved.
Licensed under the CC BY-NC-SA 4.0 license (https://creativecommons.org/licenses/by-nc-sa/4.0/).
"""

import torch
import math
import typing

#NOTE: from https://github.com/jxbz/fromage 

class Fromage(torch.optim.Optimizer):
    r"""Implements the Fromage optimisation algorithm.

        ??? info "[On the distance between two neural networks and the stability of learning](https://arxiv.org/pdf/2002.03432.pdf)"
            **Abstract**: This paper relates parameter distance to gradient breakdown for a broad class of
            nonlinear compositional functions. The analysis leads to a new distance function
            called deep relative trust and a descent lemma for neural networks. Since the resulting learning rule seems not to require learning rate grid search, it may unlock a simpler workflow for training deeper and more complex neural networks. Please find
            the Python code used in this paper here: https://github.com/jxbz/fromage.

        ??? info "[GitHub @ jxbz/fromage](https://github.com/jxbz/fromage)"

        Arguments:
            lr (float, required): The learning rate. 0.01 is a good initial value to try.
            p_bound (float, optional): Restricts the optimisation to a bounded set. A
                value of 2.0 restricts parameter norms to lie within 2x their
                initial norms. This regularises the model class.
    """
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=0.01,
        p_bound: float=None
    ):
        self.p_bound = p_bound
        defaults = dict(lr=lr)
        super(Fromage, self).__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                state = self.state[p]
                if len(state) == 0 and self.p_bound is not None:
                    state['max'] = self.p_bound*p.norm().item()
                
                d_p = p.grad.data
                d_p_norm = p.grad.norm()
                p_norm = p.norm()

                if p_norm > 0.0 and d_p_norm > 0.0:
                    p.data.add_(-group['lr'], d_p * (p_norm / d_p_norm))
                else:
                    p.data.add_(-group['lr'], d_p)
                p.data /= math.sqrt(1+group['lr']**2)

                if self.p_bound is not None:
                    p_norm = p.norm().item()
                    if p_norm > state['max']:
                        p.data *= state['max']/p_norm

        return loss


from moai.utils.arguments import (
    ensure_numeric_list,
    ensure_string_list,
)

import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import inspect
import typing
import itertools
import toolz
import logging

log = logging.getLogger(__name__)

__all__ = ["Weighted"]

from moai.monads.execution.cascade import _create_accessor

__REDUCTIONS__ = {
    'sum': torch.sum,
    'mean': torch.mean,
    'iou': lambda t: torch.mean(1.0 - (t / t[0].numel())),
}

class Weighted(torch.nn.ModuleDict):
    def __init__(self,
        losses: omegaconf.DictConfig,
        **kwargs: typing.Mapping[str, typing.Any]
    ):
        super(Weighted, self).__init__()
        self.execs, self.weights, self.reductions = [], [], []
        if not len(losses):
            log.warning("A weighted combination of losses is being used for supervising the model, but no losses have been assigned.")
        loop = ((key, params) for key, params in kwargs.items() if key in losses)
        #NOTE: check for not found keys and notify the potential error
        errors = [k for k in kwargs if k not in losses]
        if errors:
            log.error("Some losses were not found in the configuration and will be ignored!")
        self.keyz = []
        for k, p in loop:
            self.add_module(k, hyu.instantiate(getattr(losses, k)))
            # last_module = toolz.last(self.modules()) # moduledict is ordered
            last_module = self[k]
            sig = inspect.signature(last_module.forward)
            p = toolz.valmap(ensure_string_list, p)
            if 'out' not in p:
                length = len(ensure_string_list(next(iter(p.values()))))
                p['out'] = [k] if length == 1 else [f'{k}_{i}' for i in range(length)]
            if 'weight' in p:
                wgts = iter(ensure_numeric_list(p['weight']))
            else:
                log.warning(f"{k} loss has no assigned weights, automatically reverting to a weight of one (1.0).")
                wgts = itertools.cycle([1.0 / len(p['out'])])
            if 'reduction' in p:
                reduction = iter(p['reduction'])
            else:
                log.warning(f"{k} loss has no assigned reduction, automatically reverting to mean reduction.")
                reduction = itertools.cycle(['mean'])                
            #TODO: there is a bug if you pass in keys that are not bracketed ([]), i.e. as a list, even for a single arg
            for keys in zip(*list(p[prop] for prop in itertools.chain(sig.parameters, ['out']) if p.get(prop) is not None)):
                accessors = [_create_accessor(k if isinstance(k, str) else toolz.get(0, k, None)) for k in keys[:-1]]
                self.execs.append(lambda tensor_dict, 
                    acc=accessors, k=keys, p=sig.parameters.keys(), f=last_module:
                    tensor_dict.update({
                        k[-1]: f(**dict(zip(p, 
                            # list(tensor_dict[i] for i in k[:-1])
                            # list(tensor_dict.get(i, None) 
                            list(
                                a(tensor_dict) for a, i in zip(acc, k[:-1]) 
                                if i is not None or None
                            )
                        )))
                    })
                )
                self.keyz.append(keys[-1])
                self.weights.append(next(wgts)) #TODO: error if no weight has been set? or implicit 1.0 ?
                self.reductions.append(next(reduction))
            
    def forward(self,
        tensors: typing.Dict[str, torch.Tensor]
    ) -> torch.Tensor:
        device = next(toolz.take(1, 
            filter(lambda t: isinstance(t, torch.Tensor), tensors.values())
        )).device
        error = torch.tensor(0.0, dtype=torch.float32, device=device)
        per_error_map = { }
        for exe, w, k, r in zip(self.execs, self.weights, self.keyz, self.reductions):
            exe(tensors)
            # e = w * (torch.sum(tensors[k]) if r == 'sum' else torch.mean(tensors[k]))
            e = w * __REDUCTIONS__[r](tensors[k])
            per_error_map[k] = e
            error += e
        return error, per_error_map


from moai.monads.utils import expand_dims

import torch
import typing

class Znorm(torch.nn.Module):
    def __init__(self,
        dims: typing.Sequence[int]
    ):
        super(Znorm, self).__init__()
        self.dims = dims

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        std, mean = torch.std_mean(x, self.dims, keepdim=True)
        return (x - mean) / std

class MinMaxNorm(torch.nn.Module):
    def __init__(self,
        min_value: float=0.0,
        max_value: float=1.0
    ):
        super(MinMaxNorm, self).__init__()
        self.register_buffer("min", torch.scalar_tensor(self.min).float())
        self.register_buffer("max", torch.scalar_tensor(self.max).float())
        self.register_buffer("range", torch.scalar_tensor(self.max - self.min).float())

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b = x.shape[0]
        mins, _ = torch.min(x.view(b, -1), dim=1)
        maxs, _ = torch.max(x.view(b, -1), dim=1)
        scale = self.range / (maxs - mins)
        # return x.sub(mins.view(b, 1, 1, 1))\
        #     .mul(scale.view(b, 1, 1, 1))\
        #     .add(self.min)
        return torch.addcmul(
            self.min, 
            x - expand_dims(mins, x), expand_dims(scale, x)
        )            


from moai.supervision.losses.regression import L1

import torch
import math
import typing
import functools

__all__ = ["Wing"]

class Wing(L1):
    r""" Implements the generalized robust Wing loss function, including its **adaptive** and **soft** variants.

    ??? note "Wing Loss"
        <p align="center"><img src="https://render.githubusercontent.com/render/math?math=%5Chuge%5Cbegin%7Bequation%7DWing%28x%29%3D%5Cleft%5C%7B%5Cbegin%7Barray%7D%7Bll%7Dw%5Cln%28%201%20%2B%20|x|%2F%5Cepsilon%29%20%26%20%5Ctext%7Bif%7D%20|x|%20%3C%20w%20%5C%5C%20|x|%20-%20C%26%5Ctext%7Botherwise%7D%5Cend%7Barray%7D%5Cright.%5Cend%7Bequation%7D"/></p>

    ??? note "Adaptive Wing Loss"
        <p align="center"><img src="https://render.githubusercontent.com/render/math?math=%5Chuge%5Cbegin%7Bequation%7D%20%5Csmall%20AdaptiveWing%28y%2C%5C!%5Chat%7By%7D%29%5C!%20%3D%5C!%20%5Cbegin%7Bcases%7D%20%20%5Comega%5C!%20%5Cln%281%5C!%20%2B%5C!%20%20%5Cdisplaystyle%20|%5Cfrac%7By%5C!-%5C!%5Chat%7By%7D%5C!%7D%7B%5Cepsilon%7D|^%7B%5Calpha-y%7D%29%5C!%20%26%5C!%20%5Ctext%7Bif%20%7D%20|%28y%5C!-%5C!%5Chat%7By%7D%29|%5C!%20%3C%5C!%20%5Ctheta%20%20%20%5C%5C%20%20%20A|y-%5Chat%7By%7D%5C!|%20-%20C%20%26%20%5Ctext%7Botherwise%7D%20%5Cend%7Bcases%7D%20%5Cend%7Bequation%7D"/></p>

    ??? note "Soft Wing Loss"
        <p align="center"><img src="https://render.githubusercontent.com/render/math?math=%5Chuge%5Cbegin%7Bequation%7D%5Cmathrm%7BSoftWing%7D%28x%29%3D%5Cleft%5C%7B%20%5Cbegin%7Barray%7D%7Bll%7D%20%5Comega%20%5Cln%28%7B1%2B%5Cfrac%7B|x|%7D%7B%5Cepsilon%7D%7D%29%26%20%5Cmathrm%7Bif%7D%5C%20|x|%20%3C%20%5Comega%20%5C%5C%20|x|%20-%20C%20%20%26%5Cmathrm%7Botherwise%7D%20%5Cend%7Barray%7D%20%5Cright.%20%5Cend%7Bequation%7D"/></p>

    ??? cite "Papers"
        [![Paper](https://img.shields.io/static/v1?label=1711.06753&message=Wing Loss for Robust Facial Landmark Localisation with Convolutional Neural Networks&color=1c1ca2&logo=arxiv&style=plastic)](https://arxiv.org/pdf/1711.06753.pdf)

        [![Paper](https://img.shields.io/static/v1?label=1904.07399&message=Adaptive Wing Loss for Robust Face Alignment via Heatmap Regression&color=1c1ca2&logo=arxiv&style=plastic)](https://arxiv.org/pdf/1904.07399.pdf)
        
        [![Paper](https://img.shields.io/static/v1?label=2006.11697&message=Fast and Accurate: Structure Coherence Component for Face Alignment&color=1c1ca2&logo=arxiv&style=plastic)](https://arxiv.org/pdf/2006.11697.pdf)

    ???+ important "Config Entry"
        === "Main Entry"
            ```yaml
            - model/supervision/losses/regression/robust: wing
            ```
        === "Parameters"    
            ```yaml
            mode: standard # one of ['standard', 'adaptive', 'soft']
            omega: 0.05
            epsilon: 1.0
            theta: 0.05
            alpha: 2.1
            omega2: 0.02 
            ```   
        === "DML"
            ```yaml
            model:
              supervision:
                losses:
                  wing:
                    mode: standard # one of ['standard', 'adaptive', 'soft']
                    omega: 0.05
                    epsilon: 1.0
                    theta: 0.05
                    alpha: 2.1
                    omega2: 0.02 
            ```
        === "Graph"
            ```yaml
            model:
              supervision:
                losses:
                  wing:
                    gt: [groundtruth_tensor_name] # the name of the mean tensor
                    pred: [prediction_tensor_name] # the name of the variance tensor
                    out: [wing_loss] # optional, will be 'wing' if omitted
            ```

    Arguments:
        mode (str, required):
            Wing variant selector parameter, can be one of [_standard_, _adaptive_, _soft_]. 
            See the respective papers for more details, each variant uses a different parameter subset.
            **Default value: standard.**
        omega (float, optional):
            The non-negative **ω** parameter that sets the range for the non-linear part, which is
            the turning point that switches between L1 and log loss.
            **Default value: 0.05.**
        epsilon (float): 
            The curvature limiting factor **ε** for the robust Wing loss. 
            **Default value: 1.0.**
        theta (float):
            The **θ** parameter that controls the switch between linear and nonlinear part for the Adaptive Wing Loss.
            **Default value: 0.05.**
        alpha (float):
            The parameter **α** used to adapt the shape of the loss function to make it smooth at point zero for the Adaptive Wing Loss.
            **Default value: 2.1.**
        omega2 (float):
            The non-negative **ω2** parameter that controls the threshold between medium and large errors for the Soft Wing Loss.
            **Default value: 0.02.**

    !!! important
        ω2 should not set to small value because it will cause gradient vanishing problem.

    !!! important 
        Note that we should not set ω to a very small value because it makes the training of a network very unstable and causes the exploding gradient problem for very small errors.

    !!! important
        Parameter α has to be slightly larger than 2 to maintain the ideal properties required due to the normalization of the predicted variable in the range of [0, 1]. 
    
    ??? info "Repositories"        
        [![Repo2](https://github-readme-stats.vercel.app/api/pin/?username=protossw512&repo=AdaptiveWingLoss&hide_border=true&title_color=1c1ca2&show_owner=true)](https://github.com/protossw512/AdaptiveWingLoss)
        [![Repo1](https://github-readme-stats.vercel.app/api/pin/?username=elliottzheng&repo=AdaptiveWingLoss&hide_border=true&title_color=1c1ca2&show_owner=true)](https://github.com/elliottzheng/AdaptiveWingLoss)
        [![Repo3](https://github-readme-stats.vercel.app/api/pin/?username=SeungyounShin&repo=Adaptive-Wing-Loss-for-Robust-Face-Alignment-via-Heatmap-Regression&hide_border=true&title_color=1c1ca2&show_owner=true)](https://github.com/SeungyounShin/Adaptive-Wing-Loss-for-Robust-Face-Alignment-via-Heatmap-Regression)

    """
    def __init__(self,  
        mode:           str='standard', # 'standard', 'adaptive', 'soft'
        omega:          typing.Optional[float]=0.05, 
        epsilon:        float=1.0,        
        theta:          float=0.05, # variable θ as a threshold to switch between linear and nonlinear part
        alpha:          float=2.1, # Note α has to be slightly larger than 2 to maintain the ideal properties        
        omega2:         float=0.02, 
    ):
        super(Wing, self).__init__()
        self.threshold = theta if mode == 'adaptive' else omega
        Cw = omega - omega * math.log(1.0 + omega / epsilon)
        Ca = theta / epsilon
        Cs = omega - omega2 * math.log(1.0 + omega / epsilon)
        self.branch = functools.partial(self._standard_branch, 
            omega=omega, epsilon=epsilon, C=Cw)\
                if mode == 'standard' else (
            functools.partial(self._adaptive_branch,
                omega=omega, epsilon=epsilon, alpha=alpha, theta=theta, C=Ca)\
                if mode == 'adaptive' else\
            functools.partial(self._soft_branch, 
                omega=omega, epsilon=epsilon, omega2=omega2, C=Cs)
        )

    @staticmethod
    def _adaptive_branch(
        L1:             torch.Tensor,
        gt:             torch.Tensor,
        omega:          float,
        epsilon:        float,
        alpha:          float,
        theta:          float,                
        C:              float,
    ) -> torch.Tensor:
        a_minus_y = alpha - gt
        A = omega * (
            1.0 / (1.0 + C ** a_minus_y)
        ) * a_minus_y * (
            C ** (a_minus_y - 1.0)
        ) / epsilon
        C = theta * A - omega * torch.log(
            1.0 + C ** a_minus_y
        )
        return omega * torch.log(1.0 + (L1 / epsilon) ** a_minus_y), A * L1 - C        

    @staticmethod
    def _standard_branch(
        L1:             torch.Tensor,
        gt:             torch.Tensor,
        omega:          float,
        epsilon:        float,
        C:              float,
    ) -> torch.Tensor:
        return omega * torch.log(1.0 + L1 / epsilon), L1 - C

    @staticmethod
    def _soft_branch(
        L1:             torch.Tensor,
        gt:             torch.Tensor,
        omega:          float,        
        epsilon:        float,
        omega2:         float,
        C:              float,
    ) -> torch.Tensor:
        return L1, omega2 * torch.log(1 + L1 / epsilon) + C        

    def forward(self,
        pred:       torch.Tensor,
        gt:         torch.Tensor,        
        weights:    torch.Tensor=None,
        mask:       torch.Tensor=None,
    ) -> torch.Tensor:
        L1 = super(Wing, self).forward(pred=pred, gt=gt)    
        branch_A, branch_B = self.branch(L1, gt)
        wing = torch.where(
            L1 < self.threshold,
            branch_A,
            branch_B
        )
        if weights is not None:
            wing = wing * weights
        if mask is not None:
            wing = wing[mask]
        return wing


import torch

__all__ = [
    "Abs",
    "Scale",
    "Plus",
    "Minus",
    "Multiply",
    "PlusOne",
    "MinusOne",
]

class Abs(torch.nn.Module):
    def __init__(self):
        super(Abs, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.abs(x)

class Scale(torch.nn.Module):
    def __init__(self, 
        value: float
    ):
        super(Scale, self).__init__()
        self.value = value
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.value

class Plus(torch.nn.Module):
    def __init__(self):
        super(Plus, self).__init__()

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return x + y

class Minus(torch.nn.Module):
    def __init__(self):
        super(Minus, self).__init__()

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return x - y

class Multiply(torch.nn.Module):
    def __init__(self):
        super(Multiply, self).__init__()

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:        
        return x * y

class PlusOne(torch.nn.Module):
    def __init__(self):
        super(PlusOne, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # return x + torch.ones_like(x)
        return x + 1.0

class MinusOne(torch.nn.Module):
    def __init__(self):
        super(MinusOne, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # return x - torch.ones_like(x)
        return x - 1.0



import albumentations.augmentations.crops.transforms as aug
import cv2

__all__ = ["RandomSizedCrop"]

#NOTE: wrapper as to circumvent hydra/tuple issue
class RandomSizedCrop(aug.RandomSizedCrop):
    def __init__(self,
        min_height: int,
        max_height: int,
        height: int,
        width: int,
        w2h_ratio: float,
        interpolation: cv2.INTER_LINEAR,
        always_apply: bool,
        p: float=1.0
    ):
        super(RandomSizedCrop, self).__init__(
            min_max_height=(min_height, max_height),
            height=height,
            width=width,
            w2h_ratio=w2h_ratio,
            interpolation=interpolation,
            always_apply=always_apply,
            p=p
        )


import os

__all__ = ['NvidiaSMI', 'DebugCUDA']

class NvidiaSMI(object):
    def __init__(self):
        if 'PROGRAMFILES' in os.environ.keys():
            nvidia_smi_path = os.path.join(
                os.environ['PROGRAMFILES'],
                'NVIDIA Corporation',
                'NVSMI'
            )
            if nvidia_smi_path not in os.environ['PATH']:
                os.environ['PATH'] = os.environ['PATH'] + ";" + nvidia_smi_path

class DebugCUDA(object):
    def __init__(self):
        os.environ['CUDA_LAUNCH_BLOCKING'] = str(1)


from moai.supervision.losses.regression import L1

import torch

#NOTE: https://arxiv.org/pdf/2007.07627.pdf 

__all__ = ["Welsch"]

class Welsch(L1):
    r"""Implements the Welsch error function.

    """
    def __init__(self,
        v: float=0.2,
    ):
        super(Welsch, self).__init__()
        self.v_sq = v * v

    def forward(self,
        pred:       torch.Tensor,
        gt:         torch.Tensor,        
        weights:    torch.Tensor=None, # float tensor
        mask:       torch.Tensor=None, # byte tensor
    ) -> torch.Tensor:
        L1 = super(Welsch, self).forward(pred=pred, gt=gt)
        welsch = 1.0 - torch.exp(-1.0 * torch.pow(L1, 2) / 2.0 * self.v_sq)
        if weights is not None:
            welsch = welsch * weights
        if mask is not None:
            welsch = welsch[mask]
        return welsch


from moai.visualization.visdom.base import Base
from moai.utils.arguments import ensure_string_list
from moai.utils.iterators import pairwise

import torch
import visdom
import functools
import typing
import logging
import numpy as np
import cv2
import colour
import math
import toolz
from PIL import Image

log = logging.getLogger(__name__)

__all__ = ["Pose2d"]

class Pose2d(Base):
    def __init__(self,
        images:         typing.Union[str, typing.Sequence[str]],
        poses:          typing.Union[str, typing.Sequence[str]],
        gt:             typing.Union[str, typing.Sequence[str]],
        pred:           typing.Union[str, typing.Sequence[str]],
        gt_masks:       typing.Union[str, typing.Sequence[str]],
        pred_masks:     typing.Union[str, typing.Sequence[str]],
        pose_structure: typing.Union[str, typing.Sequence[str]],
        coords:         typing.Union[str, typing.Sequence[str]],
        color_gt:       typing.Union[str, typing.Sequence[str]],
        color_pred:     typing.Union[str, typing.Sequence[str]],
        name:           str="default",
        ip:             str="http://localhost",
        port:           int=8097,
        reverse_coords: bool=False,
        rotate_image:   bool=False,
        transparency:   float=0.4,
        scale:          float=1.0,
        use_mask:       bool=True,
    ):
        super(Pose2d, self).__init__(name, ip, port)
        self.images = ensure_string_list(images)
        self.poses = ensure_string_list(poses)
        self.gt = ensure_string_list(gt)
        self.pred = ensure_string_list(pred)
        self.gt_masks = ensure_string_list(gt_masks)
        self.pred_masks = ensure_string_list(pred_masks)
        self.pose_structure = ensure_string_list(pose_structure)
        self.coords = ensure_string_list(coords)
        self.color_gt = list(map(colour.web2rgb, ensure_string_list(color_gt)))
        self.color_pred = list(map(colour.web2rgb, ensure_string_list(color_pred)))
        self.reverse = reverse_coords
        self.rotate = rotate_image
        self.transparency = transparency
        self.scale = scale
        self.use_mask = use_mask
        self.viz_pose = {
            'human_pose2d': functools.partial(self.__draw_human_pose2d, 
                self.visualizer, marker=cv2.MARKER_DIAMOND, 
                rotate=self.rotate, transparency=self.transparency,
                scale=self.scale
            ),
        }
        self.xforms = { #TODO: extract these into a common module
            'ndc': lambda coord, img: torch.addcmul(
                torch.scalar_tensor(0.5).to(coord), coord, torch.scalar_tensor(0.5).to(coord)
            ) * torch.Tensor([*img.shape[2:]]).to(coord).expand_as(coord),
            'coord': lambda coord, img: coord,
            'norm': lambda coord, img: coord * torch.Tensor([*img.shape[2:]]).to(coord).expand_as(coord),
        }
        self.access = lambda td, k: toolz.get_in(k.split('.'), td)

    @property
    def name(self) -> str:
        return self.env_name
        
    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        for img, poses, gt, pred, gt_masks, pred_masks, pose_struct, gt_c, pred_c, coord in zip(
            self.images, self.poses, self.gt, self.pred, self.gt_masks, self.pred_masks,  [self.pose_structure, ],
            self.color_gt, self.color_pred, self.coords
        ):
            gt_coord = self.access(tensors, gt).detach()
            pred_coord = self.access(tensors, pred).detach()
            gt_masks = self.access(tensors, gt_masks)
            pred_masks = self.access(tensors, pred_masks)
            if self.reverse:
                gt_coord = gt_coord.flip(-1)
                pred_coord = pred_coord.flip(-1)
            image = self.access(tensors, img).detach()
            self.viz_pose[poses](
                image,
                self.xforms[coord](gt_coord, image),
                self.xforms[coord](pred_coord, image),
                gt_masks.detach() if self.use_mask else torch.ones_like(gt_coord[..., 0:1]),
                pred_masks.detach() if self.use_mask else torch.ones_like(pred_coord[..., 0:1]),
                pose_struct,
                np.uint8(np.array(list(gt_c)) * 255),
                np.uint8(np.array(list(pred_c)) * 255),
                # np.uint8(np.array(list(reversed(gt_c))) * 255),
                # np.uint8(np.array(list(reversed(pred_c))) * 255),
                coord, img, img, self.name
            )    
    
    @staticmethod
    def __draw_human_pose2d(
        visdom:             visdom.Visdom,
        images:             torch.Tensor,
        gt_coordinates:     torch.Tensor,
        pred_coordinates:   torch.Tensor,
        gt_masks:           torch.Tensor,
        pred_masks:         torch.Tensor,
        pose_structure:     typing.List[typing.List[int]],
        gt_color:           typing.List[float],
        pred_color:         typing.List[float],
        coord:              str,
        key:                str,
        win:                str,
        env:                str,
        marker:             int,
        rotate:             bool,
        transparency:       float,
        scale:              float,
    ):
        b, _, h, w = images.shape
        imgs = np.zeros([b, 3, int(scale * h), int(scale * w)], dtype=np.uint8) if not rotate \
            else np.zeros([b, 3, int(scale * w), int(scale * h)], dtype=np.uint8)
        gt_coords = gt_coordinates.cpu().int()
        pred_coords = pred_coordinates.cpu().int()
        gt_coords = torch.flip(gt_coords, dims=[-1])
        pred_coords = torch.flip(pred_coords, dims=[-1])
        gt_coords = gt_coords.numpy()
        pred_coords = pred_coords.numpy()
        diagonal = torch.norm(torch.Tensor([*imgs.shape[2:]]), p=2)
        marker_size = int(0.015 * diagonal) #TODO: extract percentage param to config?
        line_size = int(0.005 * diagonal) #TODO: extract percentage param to config?
        for i in range(imgs.shape[0]):
            img = images[i, ...].cpu().numpy().transpose(1, 2, 0) * 255.0
            img = img.copy().astype(np.uint8) if img.shape[2] > 1\
                else cv2.cvtColor(img.copy().astype(np.uint8), cv2.COLOR_GRAY2RGB)
            bg = img.copy()
            for coords, color, masks in zip(
                [gt_coords, pred_coords],
                [gt_color, pred_color],
                [gt_masks, pred_masks]
            ):       
                coord_i = coords[i, ...]
                for kpts_group in pose_structure:
                    for (a, b) in pairwise(kpts_group):
                    # for j in range(len(kpts_group) - 1):                    
                        if torch.sum(masks[i, a]) and torch.sum(masks[i, b]):
                            start_xy = tuple(coord_i[a])
                            end_xy = tuple(coord_i[b])
                            X = (start_xy[0], end_xy[0])
                            Y = (start_xy[1], end_xy[1])
                            mX = np.mean(X)
                            mY = np.mean(Y)
                            length = ((Y[0] - Y[1]) ** 2 + (X[0] - X[1]) ** 2) ** 0.5
                            angle = math.degrees(math.atan2(Y[0] - Y[1], X[0] - X[1]))
                            stickwidth = line_size
                            # polygon = cv2.ellipse2Poly(
                            #     (int(mX or 0), int(mY or 0)),
                            #     (int(length/2 or 1), int(stickwidth or 1)),
                            #     int(angle), 0, 360, 1
                            # )
                            # cv2.fillConvexPoly(bg, polygon, color.tolist())
                            cv2.line(img, start_xy, end_xy, color.tolist(), thickness=line_size)
                for k, coord in enumerate(coord_i):
                    if torch.sum(masks[i, k]):
                        if marker < 0:
                            cv2.circle(bg, 
                                tuple(coord), 
                                marker_size, color.tolist(), thickness=line_size
                            )
                        else:
                            cv2.drawMarker(bg, 
                                tuple(coord),
                                color.tolist(),
                                marker, marker_size, line_size
                            )

                img = cv2.addWeighted(bg, transparency, img, 1 - transparency, 0)
            h, w = img.shape[:2]
            if scale != 1.0:
                img = np.array(Image.fromarray(img).resize(
                    (int(w * scale), int(h * scale)), Image.ANTIALIAS
                ))
            imgs[i, ...] = img.transpose(2, 0, 1) if not rotate\
                else cv2.rotate(img, cv2.ROTATE_90_CLOCKWISE).transpose(2, 0, 1)
        visdom.images(
            imgs,
            # np.flip(imgs, axis=1),
            win=win,
            env=env,
            opts={
                'title': key,
                'caption': key,
                'jpgquality': 50,
            }
        )



import moai.utils.engine as mieng

import omegaconf.omegaconf
import logging

log = logging.getLogger(__name__)

__all__ = ["Exporters"]

class Exporters(mieng.Collection, mieng.Interval):
    def __init__(self,
        batch_interval:         int,
        exporters:              omegaconf.DictConfig,        
    ):
        mieng.Interval.__init__(self, batch_interval)
        mieng.Collection.__init__(self, 
            items=exporters, name="exporters"
        )
        log.info(f"Exporting data at a {batch_interval} step interval.")




import torch

__all__ = ["Clamp"]

class Clamp(torch.nn.Module):
    min_value: float
    max_value: float

    def __init__(self,        
        min_value: float=0.0,
        max_value: float=1.0,
    ):
        super(Clamp, self).__init__()
        self.min_value = min_value
        self.max_value = max_value

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.clamp(x, min=self.min_value, max=self.max_value)


import torch

#NOTE: https://discuss.pytorch.org/t/multi-class-cross-entropy-loss-and-softmax-in-pytorch/24920/4

__all__ = ["CrossEntropy"]

#TODO: to add class weightings
class CrossEntropy(torch.nn.CrossEntropyLoss):
    def __init__(self,
        ignore_index: int=-100,        
    ):
        super(CrossEntropy, self).__init__(            
            ignore_index=ignore_index,
            reduction='none'
        )

    def forward(self, 
        pred: torch.Tensor, # logits [B, C, ...] (i.e. raw predictions, no softmax applied -- see link above)
        gt: torch.Tensor, # class ids [B, 1, ...]        
    ) -> torch.Tensor:
        return super(CrossEntropy, self).forward(pred, gt)


from moai.supervision.losses.distribution.kl_divergence import KL
from moai.supervision.losses.distribution.std_kl import StandardNormalKL
from moai.supervision.losses.distribution.lambda_divergence import Lambda, JS
from moai.supervision.losses.distribution.variance import VarianceRegularization

__all__ = [
    "KL",
    "Lambda",
    "JS",
    "StandardNormalKL",
    "VarianceRegularization",
]


import typing
import inspect
import torch

__all__ = ["instantiate"]

def instantiate(
        which_type: typing.Type,
        **kwargs: typing.Dict
    ) -> torch.nn.Module:
        ctor_sig = inspect.signature(which_type)
        ctor_args, ctor_kwargs = { } , { }
        all_params = dict(kwargs)
        for k in ctor_sig.parameters.keys():
            if(k in all_params.keys() 
                and ctor_sig.parameters[k].default == inspect._empty
                and (ctor_sig.parameters[k].kind == inspect._ParameterKind.POSITIONAL_OR_KEYWORD
                or ctor_sig.parameters[k].kind == inspect._ParameterKind.POSITIONAL_ONLY
            )):
                ctor_args[k] = all_params[k]
            if(k in all_params.keys() 
                and ctor_sig.parameters[k].default != inspect._empty
                and (ctor_sig.parameters[k].kind == inspect._ParameterKind.POSITIONAL_OR_KEYWORD
                or ctor_sig.parameters[k].kind == inspect._ParameterKind.KEYWORD_ONLY
            )):
                ctor_kwargs[k] = all_params[k]
        bound_args = ctor_sig.bind(*ctor_args.values(), **ctor_kwargs)
        return which_type(
            *bound_args.args,
            **bound_args.kwargs
        )


import moai.networks.lightning as minet
import moai.nn.convolution as mic
import moai.nn.deconvolution as mid
import moai.nn.linear as mil

import torch
import inspect
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging
import random
import numpy
import toolz

log = logging.getLogger(__name__)

__all__ = ["VariationalAutoencoder"]

class VariationalAutoencoder(minet.FeedForward):
    def __init__(self, 
        configuration:  omegaconf.DictConfig, # in & out features, input
        preproc:        omegaconf.DictConfig,
        residual:       omegaconf.DictConfig=None,
        downscale:      omegaconf.DictConfig=None,
        data:           omegaconf.DictConfig=None,
        parameters:     omegaconf.DictConfig=None,
        feedforward:    omegaconf.DictConfig=None,        
        monads:         omegaconf.DictConfig=None,        
        supervision:    omegaconf.DictConfig=None,
        validation:     omegaconf.DictConfig=None,
        visualization:  omegaconf.DictConfig=None,
        export:         omegaconf.DictConfig=None,
    ):
        super(VariationalAutoencoder, self).__init__(
            feedforward=feedforward, monads=monads, 
            visualization=visualization, data=data,
            supervision=supervision, validation=validation,
            export=export, parameters=parameters,
        )
        self.input = configuration.input
        self.repeat_val = configuration.repeat_val
        self.traversal_len = configuration.traversal_len
        self.traversal_step = configuration.traversal_step
        self.traversal_dim = configuration.traversal_dim
        self.traversal_init_value = configuration.traversal_init_value
        self.hidden_dim = configuration.hidden_dim
        self.latent_dim = configuration.latent_dim
        in_features = configuration.in_features
        encoder_blocks = torch.nn.ModuleList()
        for dim in self.hidden_dim:
            encoder_blocks.append(mic.make_conv_block(
                block_type="conv2d",
                convolution_type=preproc.convolution.type,
                in_features=in_features, 
                out_features=dim,
                activation_type=preproc.convolution.activation.type,
                convolution_params=preproc.convolution.params
            ))
            in_features = dim
        self.linear_mu = mil.make_linear_block(
            block_type=preproc.linear.type,
            linear_type="linear",
            activation_type = preproc.linear.activation.type,
            in_features=in_features * 4 * 4,
            out_features=self.latent_dim
        )
        self.linear_logvar = mil.make_linear_block(
            block_type=preproc.linear.type,
            linear_type="linear",
            activation_type = preproc.linear.activation.type,
            in_features=in_features * 4 * 4,
            out_features=self.latent_dim
        )
        self.decoder_input = mil.make_linear_block(
            block_type=preproc.linear.type,
            linear_type="linear",
            activation_type = preproc.linear.activation.type,
            in_features=self.latent_dim,
            out_features=self.hidden_dim[-1] * 4 * 4
        )
        self.hidden_dim.reverse()
        decoder_blocks = torch.nn.ModuleList()
        for i in range(len(self.hidden_dim) - 1):
            decoder_blocks.append(mid.make_deconv_block(
                block_type="deconv2d",
                deconvolution_type=preproc.deconvolution.type,
                in_features=self.hidden_dim[i], 
                out_features=self.hidden_dim[i + 1],
                activation_type=preproc.deconvolution.activation.type,
                deconvolution_params=preproc.deconvolution.params
            ))
        decoder_blocks.append(mid.make_deconv_block(
            block_type="deconv2d",
            deconvolution_type=preproc.deconvolution.type,
            in_features=self.hidden_dim[-1], 
            out_features=self.hidden_dim[-1],
            activation_type=preproc.deconvolution.activation.type,
            deconvolution_params=preproc.deconvolution.params
            ))
        decoder_blocks.append(mic.make_conv_block(
            block_type="conv2d",
            convolution_type=preproc.convolution.type,
            in_features=self.hidden_dim[-1], 
            out_features=configuration.out_features,
            activation_type="tanh",
            convolution_params={"kernel_size": 3, "stride": 1, "padding": 1, "output_padding": 1}
            ))
        self.encoder = torch.nn.Sequential(*encoder_blocks)
        self.decoder = torch.nn.Sequential(*decoder_blocks)

    def encode(self,
        x: torch.Tensor
    ) -> [torch.Tensor, torch.Tensor]:
        x = self.encoder(x)
        x = torch.flatten(x, start_dim=1)
        mu = self.linear_mu(x)
        logvar = self.linear_logvar(x)
        return [mu, logvar]

    def decode(self,
        z: torch.Tensor
    ) -> torch.Tensor:
        x = self.decoder_input(z)
        x = x.view(-1, self.hidden_dim[0], 4, 4)
        x = self.decoder(x)
        return x

    def reparameterize(self,
        mu: torch.Tensor,
        logvar: torch.Tensor
    ) -> torch.Tensor:
        std = torch.exp(0.5*logvar)
        eps = torch.randn_like(std)
        return mu + eps*std

    def forward(self, 
        td: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        x = td[self.input]
        mu, logvar = self.encode(x)
        z = self.reparameterize(mu, logvar)
        out = self.decode(z)
        td[f"mu"] = mu
        td[f"logvar"] = logvar
        td[f"latent"] = z
        td[f"generated"] = out
        return td

    def training_step_end(self, 
        train_outputs: typing.Dict[str, typing.Union[torch.Tensor, typing.Dict[str, torch.Tensor]]]
    ) -> None:
        if self.global_step and (self.global_step % self.visualizer.interval == 0):
            self.visualizer.visualizers[0](train_outputs['tensors'])
        if self.global_step and (self.global_step % self.exporter.interval == 0):
            self.exporter(train_outputs['tensors'])
        return train_outputs['loss']

    def val_dataloader(self) -> torch.utils.data.DataLoader:
        log.info(f"Instantiating ({self.data.val.iterator._target_.split('.')[-1]}) validation set data iterator")
        validation_loader = Repeater(self.latent_dim, self.repeat_val)
        return validation_loader

    def validation_step(self,
        batch: typing.Dict[str, torch.Tensor],
        batch_nb: int
    ) -> typing.Dict[str, torch.Tensor]:
        traversal_value = self.traversal_init_value
        img_list = []
        for i in range(self.traversal_len):
            if self.traversal_dim < 0:
                batch[:] = traversal_value
            else:
                batch[self.traversal_dim] = traversal_value
            generated = self.decode(batch)
            img_list.append(generated)
            traversal_value += self.traversal_step
        generated = torch.stack(img_list, 1)
        return generated.squeeze(0)

    def validation_epoch_end(self,
        tensors: typing.Dict[str, torch.Tensor]
    ) -> None:
        self.visualizer.latent_visualizers[0](torch.stack(tensors, 1))


class Repeater(torch.utils.data.Dataset):
    def __init__(self,
        z_dim:   int,     
        repeat:  int
    ):
        super().__init__()
        self.repeat = repeat
        self.z_dim = z_dim
        self.total_z = torch.randn(self.repeat, self.z_dim)

    def __len__(self) -> int:
        return self.repeat

    def __getitem__(self, index: int) -> torch.Tensor:
        return self.total_z[index]


import moai.utils.engine as mieng

import torch
import omegaconf.omegaconf
import logging 

log = logging.getLogger(__name__)

__all__ = ["Schedulers"]

class Schedulers(mieng.Collection):
    def __init__(self, 
        schedulers: omegaconf.DictConfig
    ):
        super(Schedulers, self).__init__(items=schedulers, name="schedulers")




import torch
import json
import cv2
import numpy as np
import typing

def load_color(filenames: str, idx: int, **kwargs) -> torch.Tensor:
    img = torch.from_numpy(
            cv2.imread(filenames['rgb_filenames'][idx]).transpose(2, 0, 1)
        ).flip(dims=[0])
    return {'color': img / 255.0 }

def load_fg_mask(filenames: str, idx: int, **kwargs) -> torch.Tensor:
    img = torch.from_numpy(
            cv2.imread(filenames['mask_fg_filenames'][idx]).transpose(2, 0, 1)
        ).flip(dims=[0])
    return {'mask_fg': img[0:1, ...] / 255.0 }
    
def load_hand_mask(filenames: str, idx: int, **kwargs) -> torch.Tensor:
    img = torch.from_numpy(
            cv2.imread(filenames['mask_hand_filenames'][idx]).transpose(2, 0, 1)
        ).flip(dims=[0])
    return {'mask_hand': img[0:1, ...] / 255.0 }

def load_keypoints3D(filenames: str, idx: int, **kwargs) -> torch.Tensor:
    if idx > len(filenames['calib_filenames']):
        idx = (idx // 8) + (idx % 8)
    with open(filenames['keypoints_filenames'][idx], 'r') as fi:
        keypoints_xyz = json.load(fi)
    return {'keypoints3D': torch.from_numpy(np.array(keypoints_xyz)).transpose(1,0)}

def load_keypoints2D(filenames: str, idx: int, **kwargs) -> torch.Tensor:
    viewpoint = filenames['rgb_filenames'][idx].split('\\')[-2].split('cam')[-1]
    if idx > len(filenames['calib_filenames']):
        idx = (idx // 8) + (idx % 8)
    with open(filenames['keypoints_filenames'][idx], 'r') as fi:
        keypoints_xyz = json.load(fi)
    with open(filenames['calib_filenames'][idx], 'r') as fi:
        calib_params = json.load(fi)
    keypoints2D = projectPoints(keypoints_xyz, calib_params['K'][int(viewpoint)])
    
    return {'keypoints2D': torch.from_numpy(keypoints2D).transpose(1,0)}

def load_mano(filenames: str, idx: int, **kwargs) -> torch.Tensor:
    with open(filenames['mano_filenames'][idx], 'r') as fi:
        mano_params = json.load(fi)
    poses, betas, global_t = split_theta(mano_params)
    mano = {}
    mano['betas'] = torch.from_numpy(np.array(betas))         # 10 shape parameters
    mano['poses'] = torch.from_numpy(np.array(poses))         # 45 rotation parameters (for 15 joints) + 3 global rotation parameters
    mano['global_t'] = torch.from_numpy(np.array(global_t))   # 3 global translation parameters
    return {'mano': mano}

def projectPoints(xyz, K):
    """ Project 3D coordinates into image space. """
    xyz = np.array(xyz)
    K = np.array(K)
    uv = np.matmul(K, xyz.T).T
    return uv[:, :2] / uv[:, -1:]

def split_theta(theta):
    shapes = theta[:10]
    poses = theta[10:58]
    global_t = theta[58:61]
    return poses, shapes, global_t



from collections.abc import Callable

import typing
import torch

__all__ = ['ImageRaw']

class ImageRaw(Callable):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, 
        data:   typing.Mapping[str, typing.Any],
        device: torch.device,
    ) -> torch.Tensor:
        return {'color': torch.tensor(0.0).to(device)}


import os
import sys
import torch
import toolz
import yaml
import pathlib
import typing
import logging
import functools

from moai.data.datasets.spherical.loaders import (
    load_color,
    load_depth,
    load_normal,
    load_semantic,
    load_structure,
    load_layout,
)

log = logging.getLogger(__name__)

__all__ = ['Pano3D']

class Pano3D(torch.utils.data.Dataset):

    __LOADERS__ = {
        'color': load_color,
        'raw': load_color,
        'emission': load_color,
        'depth': load_depth,
        'normal': load_normal,
        'semantic': load_semantic,
        'structure': load_structure,
        'layout': load_layout,
        'color_up': functools.partial(load_color, position='up'),
        'color_down': functools.partial(load_color, position='down'),
        'color_left': functools.partial(load_color, position='left'),
        'color_right': functools.partial(load_color, position='right'),        
        'depth_up': functools.partial(load_depth, position='up'),
        'depth_down': functools.partial(load_depth, position='down'),
        'depth_left': functools.partial(load_depth, position='left'),
        'depth_right': functools.partial(load_depth, position='right'),
        'normal_up': functools.partial(load_normal, position='up'),
        'normal_down': functools.partial(load_normal, position='down'),
        'normal_left': functools.partial(load_normal, position='left'),
        'normal_right': functools.partial(load_normal, position='right'),
        'semantic_up': functools.partial(load_semantic, position='up'),
        'semantic_down': functools.partial(load_semantic, position='down'),
        'semantic_left': functools.partial(load_semantic, position='left'),
        'semantic_right': functools.partial(load_semantic, position='right'),
        'structure_up': functools.partial(load_structure, position='up'),
        'structure_down': functools.partial(load_structure, position='down'),
        'structure_left': functools.partial(load_structure, position='left'),
        'structure_right': functools.partial(load_structure, position='right'),
        'layout_up': functools.partial(load_layout, position='up'),
        'layout_down': functools.partial(load_layout, position='down'),
        'layout_left': functools.partial(load_layout, position='left'),
        'layout_right': functools.partial(load_layout, position='right'),
    }

    def __init__(self,
        root:       typing.Union[str, pathlib.Path],
        part:       str,
        split:      typing.Union[str, pathlib.Path],
        types:      typing.Sequence[str],
        max_depth:  float=8.0,
    ):
        folder = self._get_folder(root, part)
        self.filenames = self._get_filenames(folder, split)
        self.loaders = [
            functools.partial(Pano3D.__LOADERS__[type], 
                max_depth=max_depth
            ) for type in types
        ]

    def _get_folder(self, root: typing.Union[str, pathlib.Path], part: str) -> str:
        if not os.path.exists(root):
            log.error((
                f"Invalid Pano3D root folder ({root}), "
                "please use the folder where the Zenodo archives have been extracted to."
            ))
            sys.exit(2)
        part_path = os.path.join(root, part)
        if not os.path.exists(part_path):
            log.error((
                "Invalid Pano3D part folder, "
                f"please use the correct part folder's name inside the root directory ({root})."
            ))
            sys.exit(2)
        return part_path

    def _get_filenames(self,
        folder: str, split: typing.Union[str, pathlib.Path]
    ) -> typing.Sequence[typing.Union[str, pathlib.Path]]:
        if not os.path.exists(split):
            log.error((
                f"Pano3D split file ({split}) does not exist. "
            ))
            sys.exit(2)
        with open(split) as f:
            data_filenames = yaml.load(f, Loader=yaml.FullLoader)
        filenames = []
        for building, renders in data_filenames.items():
            for render in renders:
                filenames.append(os.path.join(folder, building, render))
        return filenames

    def __len__(self) -> int:
        return len(self.filenames)

    def __getitem__(self, index: int) -> typing.Mapping[str, torch.Tensor]:
        out = { }
        filename = self.filenames[index]
        for loader in self.loaders:
            out = toolz.merge(out, loader(filename))
        return out